﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using MyGeometry;

namespace ShapeGraph
{
	public class BoundbingBoxOptimizer
	{
		private int multivariate = 1; // number of functions
		private int nParaPerUnit = 6;
		private double scaleWeight = 10;
        private double heightWeight = 100;
        private double constWeight = 10000;
		private double groundWeight = 1000;
		private double symmWeight = 10;
		private double contactWeight = 10;
		private double[] initialx = null;


		private List<MeshComponent> meshcomponents = null;
		private List<PolyProxy> boxes = null;
		private List<PolyProxy> symmBoxes = null;
		private List<ModelGraphEdge> edges = null;

        private List<int> constNodes = null;
        private List<Vector3d> constPos = null;

		private List<int> groundNodes = null;

		private double miny = 0;

		// initialize, passing the parameters, objects, etc.
		public void Init(Subgraph sg, List<ModelGraphNode> nodes, List<Vector3d> pos)
		{
			this.edges = sg.ingoingEdges;

            this.constNodes = new List<int>();
            foreach (ModelGraphNode mn in nodes)
            {
                int index = sg.nodeSets.IndexOf(mn);
                this.constNodes.Add(index);
            }
            this.constPos = pos;
			//	this.contacts.Clear();

			this.meshcomponents = new List<MeshComponent>();
			this.boxes = new List<PolyProxy>();
			List<double> xx = new List<double>();
			foreach (ModelGraphNode mn in sg.nodeSets)
			{
                MeshComponent mc = mn.meshComponent;
                if (mc.boundingBox != null)
				{
					this.meshcomponents.Add(mc);
					this.boxes.Add(mc.boundingBox);
                    xx.AddRange(mc.boundingBox.center.pos.ToArray());
                    mc.boundingBox.scale = mc.boundingBox.points3[6].pos 
                        - mc.boundingBox.points3[0].pos;
                    xx.AddRange(mc.boundingBox.scale.ToArray());
				}
			}

			// get symmetry constraints
			this.symmBoxes = new List<PolyProxy>();
            this.multivariate = this.boxes.Count * nParaPerUnit + sg.ingoingEdges.Count * 3
                + this.constPos.Count * 3;//
            //+ (this.symmBoxes.Count / 2) * 7; // # of functions
			this.initialx = xx.ToArray();

		}

        public void Init(ModelGraph g, List<ModelGraphNode> nodes, List<Vector3d> pos)
        {
            this.edges = g.graphEdges;

            this.constNodes = new List<int>();
            foreach (ModelGraphNode mn in nodes)
            {
                int index = g.graphNodes.IndexOf(mn);
                this.constNodes.Add(index);
            }
            this.constPos = pos;
            //	this.contacts.Clear();

            this.meshcomponents = new List<MeshComponent>();
            this.boxes = new List<PolyProxy>();
            List<double> xx = new List<double>();

			groundNodes = new List<int>();

            for(int i = 0; i < g.graphNodes.Count;++i) 
            {
				ModelGraphNode mn = g.graphNodes[i];
                MeshComponent mc = mn.meshComponent;
                if (mc.boundingBox != null)
                {
                    this.meshcomponents.Add(mc);
                    this.boxes.Add(mc.boundingBox);
                    xx.AddRange(mc.boundingBox.center.pos.ToArray());
                    mc.boundingBox.scale = mc.boundingBox.points3[6].pos
                        - mc.boundingBox.points3[0].pos;
                    xx.AddRange(mc.boundingBox.scale.ToArray());
                }
				if (mn.isgroundcontact) groundNodes.Add(i);
            }

            // get symmetry constraints
            this.symmBoxes = new List<PolyProxy>();
            //foreach (MeshComponent node in g.meshComponents)
            //{
            //    if (node.symmetry != null && !this.symmBoxes.Contains(node.meshComponent.boundingBox))
            //    {
            //        this.symmBoxes.Add(node.meshComponent.boundingBox);
            //        this.symmBoxes.Add(node.symmetry.meshComponent.boundingBox);
            //    }
            //}


            this.multivariate = this.boxes.Count * nParaPerUnit + g.graphEdges.Count * 3
                + this.constPos.Count * 3 +this.groundNodes.Count;//
            //+ (this.symmBoxes.Count / 2) * 7; // # of functions
            this.initialx = xx.ToArray();

        }

		// function and jacobi matrix for (LM)
		private void function(double[] x, double[] funcs, object obj)
		{
			int J = 0; // function index
			for (J = 0; J < this.initialx.Length; J += nParaPerUnit)
			{
				funcs[J] = x[J] - this.initialx[J];
				funcs[J + 1] = x[J + 1] - this.initialx[J + 1];
				funcs[J + 2] = x[J + 2] - this.initialx[J + 2];
				funcs[J + 3] = scaleWeight * (x[J + 3] - this.initialx[J + 3]);
				funcs[J + 4] = heightWeight * (x[J + 4] - this.initialx[J + 4]);
				funcs[J + 5] = scaleWeight * (x[J + 5] - this.initialx[J + 5]);
			}

            for(int i = 0; i < this.constNodes.Count;++i)
            {
                int nidx = this.constNodes[i] * nParaPerUnit;
                funcs[J] = constWeight * (x[nidx] - this.constPos[i][0]);
                funcs[J + 1] = constWeight * (x[nidx + 1] - this.constPos[i][1]);
                funcs[J + 2] = constWeight * (x[nidx + 2] - this.constPos[i][2]);
                J += 3;
            }

			for (int i = 0; i < this.groundNodes.Count; ++i)
			{
				int nidx = this.groundNodes[i] * nParaPerUnit;
				funcs[J] = groundWeight * (x[nidx + 1] - miny);
				J += 1;
			}
            
			// symmetry constraints
			for (int i = 0; i < this.symmBoxes.Count; i += 2)
			{
				PolyProxy _box1 = this.symmBoxes[i];
				PolyProxy _box2 = this.symmBoxes[i + 1];
				int _index1 = this.boxes.IndexOf(_box1) * nParaPerUnit;
				int _index2 = this.boxes.IndexOf(_box2) * nParaPerUnit;

				Vector3d o = (_box1.center.pos + _box2.center.pos) / 2;			// reflectional symmetry plane center -- to be defined
				Vector3d n = (_box1.center.pos - _box2.center.pos).Normalize();	// reflectional symmetry plane normal

				// centers
				// 1. n*(c_i+c_j) = 2o*n
				funcs[J] = symmWeight * (n.Dot(new Vector3d(x, _index1) + new Vector3d(x, _index2)) - 2 * n.Dot(o));
				J++;
				// 2. nx(c_i-c_j) = 0
				funcs[J] = symmWeight * ((x[_index1] - x[_index2]) * n.z - (x[_index1 + 2] - x[_index2 + 2]) * n.x);
				J++;
				funcs[J] = symmWeight * ((x[_index1 + 1] - x[_index2 + 1]) * n.x - (x[_index1] - x[_index2]) * n.y);
				J++;
				funcs[J] = symmWeight * ((x[_index1 + 2] - x[_index2 + 2]) * n.y - (x[_index1 + 1] - x[_index2 + 1]) * n.z);
				J++;

				// scales
				funcs[J] = symmWeight * (x[_index1 + 3] - x[_index2 + 3]);
				J++;
				funcs[J] = symmWeight * (x[_index1 + 4] - x[_index2 + 4]);
				J++;
				funcs[J] = symmWeight * (x[_index1 + 5] - x[_index2 + 5]);
				J++;
			}

			// contacts constraints
			foreach (ModelGraphEdge e in this.edges)
			{
				PolyProxy _box1 = e.conntectingNodes[0].meshComponent.boundingBox;
				PolyProxy _box2 = e.conntectingNodes[1].meshComponent.boundingBox;
				Vector3d c1 = _box1.frame.GetPointLocalCoord(e.pos);
				Vector3d c2 = _box2.frame.GetPointLocalCoord(e.pos);
				Vector3d s1 = _box1.scale;
				Vector3d s2 = _box2.scale;
				c1.x /= s1.x; c1.y /= s1.y; c1.z /= s1.z;
				c2.x /= s2.x; c2.y /= s2.y; c2.z /= s2.z;

				int _index1 = this.boxes.IndexOf(_box1) * nParaPerUnit;
				int _index2 = this.boxes.IndexOf(_box2) * nParaPerUnit;

				// c_i + [x_i|y_i|z_i]*c1 - cj - [x_j|y_j|z_j]*c2 = 0;
				funcs[J] = contactWeight * (x[_index1] - x[_index2]
					+ c1.x * _box1.frame.x.x * x[_index1 + 3] - c2.x * _box2.frame.x.x * x[_index2 + 3]
					+ c1.y * _box1.frame.y.x * x[_index1 + 4] - c2.y * _box2.frame.y.x * x[_index2 + 4]
					+ c1.z * _box1.frame.z.x * x[_index1 + 5] - c2.z * _box2.frame.z.x * x[_index2 + 5]
					);
				J++;
				funcs[J] = contactWeight * (x[_index1 + 1] - x[_index2 + 1]
					+ c1.x * _box1.frame.x.y * x[_index1 + 3] - c2.x * _box2.frame.x.y * x[_index2 + 3]
					+ c1.y * _box1.frame.y.y * x[_index1 + 4] - c2.y * _box2.frame.y.y * x[_index2 + 4]
					+ c1.z * _box1.frame.z.y * x[_index1 + 5] - c2.z * _box2.frame.z.y * x[_index2 + 5]
					);
				J++;
				funcs[J] = contactWeight * (x[_index1 + 2] - x[_index2 + 2]
					+ c1.x * _box1.frame.x.z * x[_index1 + 3] - c2.x * _box2.frame.x.z * x[_index2 + 3]
					+ c1.y * _box1.frame.y.z * x[_index1 + 4] - c2.y * _box2.frame.y.z * x[_index2 + 4]
					+ c1.z * _box1.frame.z.z * x[_index1 + 5] - c2.z * _box2.frame.z.z * x[_index2 + 5]
					);
				J++;
			}

			
		}


		private void jacobi(double[] x, double[] funcs, double[,] jac, object obj)
		{
			int J = 0; // function index
			for (J = 0; J < this.initialx.Length; ++J)
			{
				funcs[J] = x[J] - this.initialx[J];
				jac[J, J] = 1.0;
			}

            for (int i = 0; i < this.constNodes.Count; ++i)
            {
                int nidx = this.constNodes[i] * nParaPerUnit;
                funcs[J] = constWeight * (x[nidx] - this.constPos[i][0]);
                funcs[J + 1] = constWeight * (x[nidx + 1] - this.constPos[i][1]);
                funcs[J + 2] = constWeight * (x[nidx + 2] - this.constPos[i][2]);
                jac[J, nidx] = constWeight;
                jac[J + 1, nidx + 1] = constWeight;
                jac[J + 2, nidx + 2] = constWeight;
                J += 3;
            }

			for (int i = 0; i < this.groundNodes.Count; ++i)
			{
				int nidx = this.groundNodes[i] * nParaPerUnit;
				funcs[J] = groundWeight * (x[nidx+1] -  miny);
				jac[J, nidx] = groundWeight;
				J += 1;
			}

			// symmetry constraints
			for (int i = 0; i < this.symmBoxes.Count; i += 2)
			{
				PolyProxy _box1 = this.symmBoxes[i];
				PolyProxy _box2 = this.symmBoxes[i + 1];
				int _index1 = this.boxes.IndexOf(_box1) * nParaPerUnit;
				int _index2 = this.boxes.IndexOf(_box2) * nParaPerUnit;

				Vector3d o = (_box1.center.pos + _box2.center.pos) / 2;			// reflectional symmetry plane center -- to be defined
				Vector3d n = (_box1.center.pos - _box2.center.pos).Normalize();	// reflectional symmetry plane normal

				// centers
				// 1. n*(c_i+c_j) = 2o*n
				funcs[J] = symmWeight * (n.Dot(new Vector3d(x, _index1) + new Vector3d(x, _index2)) - 2 * n.Dot(o));
				jac[J, _index1] = symmWeight * n.x; jac[J, _index1 + 1] = symmWeight * n.y; jac[J, _index1 + 2] = symmWeight * n.z;
				jac[J, _index2] = symmWeight * n.x; jac[J, _index2 + 1] = symmWeight * n.y; jac[J, _index2 + 2] = symmWeight * n.z;
				J++;
				// 2. nx(c_i-c_j) = 0
				funcs[J] = symmWeight * ((x[_index1] - x[_index2]) * n.z - (x[_index1 + 2] - x[_index2 + 2]) * n.x);
				jac[J, _index1] = symmWeight * n.z; jac[J, _index2] = -symmWeight * n.z;
				jac[J, _index1 + 2] = -symmWeight * n.x; jac[J, _index2 + 2] = symmWeight * n.x;
				J++;
				funcs[J] = symmWeight * ((x[_index1 + 1] - x[_index2 + 1]) * n.x - (x[_index1] - x[_index2]) * n.y);
				jac[J, _index1 + 1] = symmWeight * n.x; jac[J, _index2 + 1] = -symmWeight * n.x;
				jac[J, _index1] = -symmWeight * n.y; jac[J, _index2] = symmWeight * n.y;
				J++;
				funcs[J] = symmWeight * ((x[_index1 + 2] - x[_index2 + 2]) * n.y - (x[_index1 + 1] - x[_index2 + 1]) * n.z);
				jac[J, _index1 + 2] = symmWeight * n.y; jac[J, _index2 + 2] = -symmWeight * n.y;
				jac[J, _index1 + 1] = -symmWeight * n.z; jac[J, _index2 + 1] = symmWeight * n.z;
				J++;

				// scales
				funcs[J] = symmWeight * (x[_index1 + 3] - x[_index2 + 3]);
				jac[J, _index1 + 3] = symmWeight;
				jac[J, _index2 + 3] = -symmWeight;
				J++;
				funcs[J] = symmWeight * (x[_index1 + 4] - x[_index2 + 4]);
				jac[J, _index1 + 4] = symmWeight;
				jac[J, _index2 + 4] = -symmWeight;
				J++;
				funcs[J] = symmWeight * (x[_index1 + 5] - x[_index2 + 5]);
				jac[J, _index1 + 5] = symmWeight;
				jac[J, _index2 + 5] = -symmWeight;
				J++;
			}

			// contacts constraints
			foreach (ModelGraphEdge e in this.edges)
			{
				PolyProxy _box1 = e.conntectingNodes[0].meshComponent.boundingBox;
				PolyProxy _box2 = e.conntectingNodes[1].meshComponent.boundingBox;
                e.pos = (_box1.center.pos + _box2.center.pos) / 2;
				Vector3d c1 = _box1.frame.GetPointLocalCoord(e.pos);
				Vector3d c2 = _box2.frame.GetPointLocalCoord(e.pos);
				Vector3d s1 = _box1.scale;
				Vector3d s2 = _box2.scale;
				c1.x /= s1.x; c1.y /= s1.y; c1.z /= s1.z;
				c2.x /= s2.x; c2.y /= s2.y; c2.z /= s2.z;

				int _index1 = this.boxes.IndexOf(_box1) * nParaPerUnit;
				int _index2 = this.boxes.IndexOf(_box2) * nParaPerUnit;

				// c_i + [x_i|y_i|z_i]*c1 - cj - [x_j|y_j|z_j]*c2 = 0;
				funcs[J] = contactWeight * (x[_index1] - x[_index2]
					+ c1.x * _box1.frame.x.x * x[_index1 + 3] - c2.x * _box2.frame.x.x * x[_index2 + 3]
					+ c1.y * _box1.frame.y.x * x[_index1 + 4] - c2.y * _box2.frame.y.x * x[_index2 + 4]
					+ c1.z * _box1.frame.z.x * x[_index1 + 5] - c2.z * _box2.frame.z.x * x[_index2 + 5]
					);
				jac[J, _index1] = contactWeight;
				jac[J, _index2] = -contactWeight;
				jac[J, _index1 + 3] = contactWeight * c1.x * _box1.frame.x.x;
				jac[J, _index2 + 3] = -contactWeight * c2.x * _box2.frame.x.x;
				jac[J, _index1 + 4] = contactWeight * c1.y * _box1.frame.y.x;
				jac[J, _index2 + 4] = -contactWeight * c2.y * _box2.frame.y.x;
				jac[J, _index1 + 5] = contactWeight * c1.z * _box1.frame.z.x;
				jac[J, _index2 + 5] = -contactWeight * c2.z * _box2.frame.z.x;
				J++;

				funcs[J] = contactWeight * (x[_index1 + 1] - x[_index2 + 1]
					+ c1.x * _box1.frame.x.y * x[_index1 + 3] - c2.x * _box2.frame.x.y * x[_index2 + 3]
					+ c1.y * _box1.frame.y.y * x[_index1 + 4] - c2.y * _box2.frame.y.y * x[_index2 + 4]
					+ c1.z * _box1.frame.z.y * x[_index1 + 5] - c2.z * _box2.frame.z.y * x[_index2 + 5]
					);
				jac[J, _index1 + 1] = contactWeight;
				jac[J, _index2 + 1] = -contactWeight;
				jac[J, _index1 + 3] = contactWeight * c1.x * _box1.frame.x.y;
				jac[J, _index2 + 3] = -contactWeight * c2.x * _box2.frame.x.y;
				jac[J, _index1 + 4] = contactWeight * c1.y * _box1.frame.y.y;
				jac[J, _index2 + 4] = -contactWeight * c2.y * _box2.frame.y.y;
				jac[J, _index1 + 5] = contactWeight * c1.z * _box1.frame.z.y;
				jac[J, _index2 + 5] = -contactWeight * c2.z * _box2.frame.z.y;
				J++;

				funcs[J] = contactWeight * (x[_index1 + 2] - x[_index2 + 2]
					+ c1.x * _box1.frame.x.z * x[_index1 + 3] - c2.x * _box2.frame.x.z * x[_index2 + 3]
					+ c1.y * _box1.frame.y.z * x[_index1 + 4] - c2.y * _box2.frame.y.z * x[_index2 + 4]
					+ c1.z * _box1.frame.z.z * x[_index1 + 5] - c2.z * _box2.frame.z.z * x[_index2 + 5]
					);
				jac[J, _index1 + 2] = contactWeight;
				jac[J, _index2 + 2] = -contactWeight;
				jac[J, _index1 + 3] = contactWeight * c1.x * _box1.frame.x.z;
				jac[J, _index2 + 3] = -contactWeight * c2.x * _box2.frame.x.z;
				jac[J, _index1 + 4] = contactWeight * c1.y * _box1.frame.y.z;
				jac[J, _index2 + 4] = -contactWeight * c2.y * _box2.frame.y.z;
				jac[J, _index1 + 5] = contactWeight * c1.z * _box1.frame.z.z;
				jac[J, _index2 + 5] = -contactWeight * c2.z * _box2.frame.z.z;
				J++;
			}

			
		}


		public double[] LMOptimize(bool useJacobi)
		{
			//
			// This is the minimization of a nonlinear function 
			// F(x0,x1) = \sum_i{fi}^2
			// using "VJ" mode of the Levenberg-Marquardt optimizer.
			// -- Optimization algorithm uses: --
			// * function vector f[] = {...,f_i,...}
			// * Jacobian matrix J = {dfi/dxj}.
			//
			double[] x = this.initialx.Clone() as double[];
			double epsg = 0.0000000001;
			double epsf = 0;
			double epsx = 0;
			int maxits = 10000;

			// first run
			alglib.minlmstate state;
			alglib.minlmreport rep;

			if (!useJacobi)
			{
				// without jacobi
				alglib.minlmcreatev(this.multivariate, x, 0.0001, out state);
				alglib.minlmsetcond(state, epsg, epsf, epsx, maxits);
				alglib.minlmoptimize(state, function, null, null);
				alglib.minlmresults(state, out x, out rep);
			}
			else
			{
				// with jacobi
				alglib.minlmcreatevj(this.multivariate, x, out state);
				alglib.minlmsetcond(state, epsg, epsf, epsx, maxits);
				alglib.minlmoptimize(state, function, jacobi, null, null);
				alglib.minlmresults(state, out x, out rep);
			}

			// update the boxes
			int index = 0;
			for (int i = 0; i < x.Length; i += nParaPerUnit)
			{
				PolyProxy oldBox = this.boxes[index];
				PolyProxy _box = new PolyProxy(
						new Vector3d(x, i),
						new Vector3d(x, i + 3),
						oldBox.frame.x,
						oldBox.frame.y,
						oldBox.frame.z
					);
				_box.index = this.boxes[index].index;
				this.boxes[index] = _box;
				this.meshcomponents[index].boundingBox = _box;
				index++;
			}


			return x;
		}
	}
}
