﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MyGeometry
{
	public class Element2D
	{
		public Element2D() { }
		public int nSamples = 16;
		public Vector3d[] points = null;
		public void Transform(Matrix4d T)
		{
			for (int i = 0, j = 0; i < nSamples; i++, j += 3)
			{
				Vector3d v = this.points[i];
				this.points[i] =
					(T * new Vector4d(v, 1)).XYZ();
			}
		}
	}
	public class Circle : Element2D
	{
		public double radius = 1.0;
		public Circle()
		{
			this.nSamples = 16;
			this.Create();
		}
		public Circle(int n)
		{
			this.nSamples = n;
			this.Create();
		}
		public void Create()
		{
			this.points = new Vector3d[nSamples];

			double theta = 2 * Math.PI / nSamples;
			for (int i = 0; i < nSamples; ++i)
			{
				double t = theta * i;
				double x = radius*Math.Cos(t);
				double y = radius*Math.Sin(t);
				this.points[i] = new Vector3d(
					x, y, 0
				);
			}
		}
	}
	public class MyRectangle : Element2D
	{
		public int level = 8;
		public double width = 1.0;
		public MyRectangle()
		{
			this.Create();
		}
		public MyRectangle(int nPointsPerEdge)
		{
			this.level = nPointsPerEdge;
			this.Create();
		}
		public void Create()
		{
			this.nSamples = this.level * 4; ;
			this.width = 1.0;

			this.points = new Vector3d[nSamples];
			
			Vector2d[] x = new Vector2d[4] {
				new Vector2d(-width, -width),
				new Vector2d(width, -width),
				new Vector2d(width, width),
				new Vector2d(-width, width)
			};
			double[] r = new double[level];
			for (int i = 0; i < level; ++i)
			{
				r[i] = (double)i / level;
			}
			Vector2d[][] p = new Vector2d[4][];
			for (int i = 0; i < 4; ++i)
			{
				p[i] = new Vector2d[level];
			}
			for (int i = 0; i < nSamples; i++)
			{
				int k = i / level;
				int l = i % level;
				int t = (k + 1) % 4;
				// x,y coords
				p[k][l] = x[k] * (1 - r[l]) + x[t] * r[l];
			}

			for (int j = 0; j < nSamples; j++)
			{
				int k = j / level;
				int l = j % level;

				this.points[j] = new Vector3d(
					p[k][l].x,
					p[k][l].y,
					0
				);
			}
		}
	}

	public class MyElement
	{
		public static Circle myCircle = new Circle();
		public static MyRectangle myRect = new MyRectangle();
		public MyElement()
		{ }
	}

	public class Vertex
	{
		public Vertex()
		{
			Pos = new Vector3d();
			AdjV = new ArrayList();
		}
		public Vertex(double x, double y, double z)
		{
			Pos = new Vector3d(x, y, z);
			AdjV = new ArrayList();
		}
		public Vertex(Vector3d v)
		{
			Pos = new Vector3d(v);
			AdjV = new ArrayList();
		}

		public int Index { get; set; }
		public Vector3d Pos { get; set; }
		public ArrayList AdjV { get; set; }
	}
	public class Face
	{
		public Face()
		{
			vList = new Vertex[3];
		}

		public int Index { get; set; }
		public Vertex[] vList { get; set; }
	}
	public class Polyhedron
	{
		public Polyhedron()
		{
			vList = new Vertex[0];
			fList = new Face[0];
		}

		public Mesh GetMesh()
		{
			ArrayList vlist = new ArrayList();
			ArrayList flist = new ArrayList();

			for (int i = 0; i < vList.Count(); i++)
			{
				for (int d = 0; d < 3; d++)
					vlist.Add(vList[i].Pos[d]);
			}
			for (int i = 0; i < fList.Count(); i++)
			{
				for (int d = 0; d < 3; d++)
					flist.Add(fList[i].vList[d].Index);
			}


			Mesh Mesh = new Mesh(vlist, flist);

			return Mesh;
		}

		public Vector3d[] GetLines()
		{
			Vector3d[] Lines = new Vector3d[lines.Count()];
			for (int i = 0; i < lines.Count(); i++)
				Lines[i] = lines[i];
			return Lines;
		}

		public Vertex[] vList;
		public Face[] fList;
		public List<int> keyPoints = new List<int>();

		public Vector3d[] lines = new Vector3d[0];
	}
	public class Bar : Polyhedron
	{
		public Bar(int level)
		{
			ArrayList VertexList = new ArrayList();
			ArrayList FaceList = new ArrayList();

			int numrow = 3 * level;
			int numsec = 4 * level;

			double lenrow = 1.0 / numrow;
			double delta = 1.0 / level;

			for (int j = 0; j < numrow + 1; j++)
			{
				for (int i = 0; i < level; i++) VertexList.Add(new Vertex(0.5 - i * delta, 0.5, j * lenrow));
				for (int i = 0; i < level; i++) VertexList.Add(new Vertex(-0.5, 0.5 - i * delta, j * lenrow));
				for (int i = 0; i < level; i++) VertexList.Add(new Vertex(-0.5 + i * delta, -0.5, j * lenrow));
				for (int i = 0; i < level; i++) VertexList.Add(new Vertex(0.5, -0.5 + i * delta, j * lenrow));
			}

			for (int i = 0; i < VertexList.Count; i++)
			{
				((Vertex)VertexList[i]).Index = i;
			}

			for (int i = 0; i < numrow; i++)
			{
				for (int j = 0; j < numsec; j++)
				{
					FaceList.Add(new Face());
					FaceList.Add(new Face());

					((Face)FaceList[2 * (i * numsec + j)]).vList[0] = (Vertex)VertexList[i * numsec + j];
					((Face)FaceList[2 * (i * numsec + j)]).vList[1] = (Vertex)VertexList[(i + 1) * numsec + (j + 1) % numsec];
					((Face)FaceList[2 * (i * numsec + j)]).vList[2] = (Vertex)VertexList[(i + 1) * numsec + j];


					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[0] = (Vertex)VertexList[i * numsec + j];
					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[1] = (Vertex)VertexList[i * numsec + (j + 1) % numsec];
					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[2] = (Vertex)VertexList[(i + 1) * numsec + (j + 1) % numsec];
				}
			}

			vList = new Vertex[VertexList.Count];
			fList = new Face[FaceList.Count];

			for (int i = 0; i < VertexList.Count; i++) vList[i] = (Vertex)VertexList[i];
			for (int i = 0; i < FaceList.Count; i++) fList[i] = (Face)FaceList[i];
		}
	}
	public class Cylinder : Polyhedron
	{
		public int numrow = 0;
		public int numsec = 0;
		public Cylinder(int level)
		{
			ArrayList VertexList = new ArrayList();
			ArrayList FaceList = new ArrayList();

			this.numrow = 3 * level;
			this.numsec = MyElement.myCircle.nSamples;

			double lenrow = 1.0 / numrow;
			double radius = MyElement.myCircle.radius;
			double angle = 2 * Math.PI / numsec;

			double z_bottom = -0.5;

			int indexer = 0;
			int ncircles = numrow + 1;
			for (int i = 0; i < ncircles; i++)
			{
				// positions
				for (int j = 0; j < numsec; j++)
				{
					double x = MyElement.myCircle.points[j].x;
					double y = MyElement.myCircle.points[j].y;
				//	VertexList.Add(new Vertex(radius * Math.Cos(j * angle), radius * Math.Sin(j * angle), z_bottom + i * lenrow));
					VertexList.Add(new Vertex(x,y, z_bottom + i * lenrow));
					indexer++;
				}
				// key points
				if (i == 0 || i == numrow)
				{
					int id = indexer - numsec;
					for (int j = 0; j < numsec; ++j)
					{
						keyPoints.Add(id++);
					}
				}
			}

			VertexList.Add(new Vertex(0, 0, z_bottom));
			VertexList.Add(new Vertex(0, 0, z_bottom+1.0));

			for (int i = 0; i < VertexList.Count; i++)
			{
				((Vertex)VertexList[i]).Index = i;
			}

			for (int i = 0; i < numrow; i++)
			{
				for (int j = 0; j < numsec; j++)
				{
					FaceList.Add(new Face());
					FaceList.Add(new Face());

					((Face)FaceList[2 * (i * numsec + j)]).vList[0] = (Vertex)VertexList[i * numsec + j];
					((Face)FaceList[2 * (i * numsec + j)]).vList[1] = (Vertex)VertexList[(i + 1) * numsec + (j + 1) % numsec];
					((Face)FaceList[2 * (i * numsec + j)]).vList[2] = (Vertex)VertexList[(i + 1) * numsec + j];

					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[0] = (Vertex)VertexList[i * numsec + j];
					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[1] = (Vertex)VertexList[i * numsec + (j + 1) % numsec];
					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[2] = (Vertex)VertexList[(i + 1) * numsec + (j + 1) % numsec];
				}
			}

			for (int j = 0; j < numsec; j++)
			{
				FaceList.Add(new Face());
				((Face)FaceList[FaceList.Count - 1]).vList[0] = (Vertex)VertexList[(j + 1) % numsec];
				((Face)FaceList[FaceList.Count - 1]).vList[1] = (Vertex)VertexList[j];
				((Face)FaceList[FaceList.Count - 1]).vList[2] = (Vertex)VertexList[VertexList.Count - 2];
			}

			for (int j = 0; j < numsec; j++)
			{
				FaceList.Add(new Face());
				((Face)FaceList[FaceList.Count - 1]).vList[0] = (Vertex)VertexList[VertexList.Count - 1];
				((Face)FaceList[FaceList.Count - 1]).vList[1] = (Vertex)VertexList[numrow * numsec + j];
				((Face)FaceList[FaceList.Count - 1]).vList[2] = (Vertex)VertexList[numrow * numsec + (j + 1) % numsec];
			}

			vList = new Vertex[VertexList.Count];
			fList = new Face[FaceList.Count];

			for (int i = 0; i < VertexList.Count; i++) vList[i] = (Vertex)VertexList[i];
			for (int i = 0; i < FaceList.Count; i++) fList[i] = (Face)FaceList[i];

			int num = 32;
			lines = new Vector3d[2 * (num * 3 + 4)];

			double a = 2 * Math.PI / num;

			for (int i = 0; i < num; i++)
			{
				lines[2 * (0 * num + i)] = new Vector3d(Math.Cos(i * a), Math.Sin(i * a), 0);
				lines[2 * (0 * num + i) + 1] = new Vector3d(Math.Cos((i + 1) % num * a), Math.Sin((i + 1) % num * a), 0);
			}
			for (int i = 0; i < num; i++)
			{
				lines[2 * (1 * num + i)] = new Vector3d(Math.Cos(i * a), Math.Sin(i * a), 0.5);
				lines[2 * (1 * num + i) + 1] = new Vector3d(Math.Cos((i + 1) % num * a), Math.Sin((i + 1) % num * a), 0.5);
			}
			for (int i = 0; i < num; i++)
			{
				lines[2 * (2 * num + i)] = new Vector3d(Math.Cos(i * a), Math.Sin(i * a), 1.0);
				lines[2 * (2 * num + i) + 1] = new Vector3d(Math.Cos((i + 1) % num * a), Math.Sin((i + 1) % num * a), 1.0);
			}

			lines[2 * (3 * num + 0)] = new Vector3d(1, 0, 0);
			lines[2 * (3 * num + 0) + 1] = new Vector3d(1, 0, 1);
			lines[2 * (3 * num + 1)] = new Vector3d(0, 1, 0);
			lines[2 * (3 * num + 1) + 1] = new Vector3d(0, 1, 1);
			lines[2 * (3 * num + 2)] = new Vector3d(-1, 0, 0);
			lines[2 * (3 * num + 2) + 1] = new Vector3d(-1, 0, 1);
			lines[2 * (3 * num + 3)] = new Vector3d(0, -1, 0);
			lines[2 * (3 * num + 3) + 1] = new Vector3d(0, -1, 1);
		}
	}
	public class Cuboid : Polyhedron
	{
		public int numrow = 0;
		public int numsec = 0;
		public List<int> xSideKeyPoints = new List<int>();
		public List<int> ySideKeyPoints = new List<int>();
		public List<int> zSideKeyPoints = new List<int>();
		public Cuboid(int level)
		{
			ArrayList VertexList = new ArrayList();
			ArrayList FaceList = new ArrayList();

			numrow = 3 * level; // divid the z-axis
			numsec = 4 * level; // divid the circles

			double lenrow = 1.0 / numrow;
			double radius = 1;
			double angle = 2 * Math.PI / numsec;

			double z_bottom = -0.5;

			int indexer = 0; double r = radius; double shift = -Math.PI / 4;
			int ncircles = numrow+1; // number of circles
			for (int i = 0; i < ncircles; i++)
			{
				for (int j = 0; j < numsec; j++)
				{
					// compute the r length
					double theta = angle * j + shift;
					double oldTheta = theta;
					int jj = (int)(theta * 4 / Math.PI);
					int kk = (jj / 2) * 2;
					theta = theta - kk * Math.PI / 4;
					theta = theta > Math.PI / 4 ? Math.PI / 2 - theta : theta;
					r = radius / Math.Cos(theta);
					
					// compute x,y coords
					double x = r * Math.Cos(oldTheta), y = r*Math.Sin(oldTheta);
					VertexList.Add(new Vertex(x, y, z_bottom + i * lenrow));
					indexer++;
				}
				// key points
				if (i == 0 || i == numrow)
				{
					int id = indexer - numsec;
					for (int j = 0; j < numsec; ++j)
					{
						this.zSideKeyPoints.Add(id++);
					}
				}
			}

			VertexList.Add(new Vertex(0, 0, z_bottom));
			VertexList.Add(new Vertex(0, 0, z_bottom+1.0));

			for (int i = 0; i < VertexList.Count; i++)
			{
				((Vertex)VertexList[i]).Index = i;
			}

			for (int i = 0; i < numrow; i++)
			{
				for (int j = 0; j < numsec; j++)
				{
					FaceList.Add(new Face());
					FaceList.Add(new Face());

					((Face)FaceList[2 * (i * numsec + j)]).vList[0] = (Vertex)VertexList[i * numsec + j];
					((Face)FaceList[2 * (i * numsec + j)]).vList[1] = (Vertex)VertexList[(i + 1) * numsec + (j + 1) % numsec];
					((Face)FaceList[2 * (i * numsec + j)]).vList[2] = (Vertex)VertexList[(i + 1) * numsec + j];

					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[0] = (Vertex)VertexList[i * numsec + j];
					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[1] = (Vertex)VertexList[i * numsec + (j + 1) % numsec];
					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[2] = (Vertex)VertexList[(i + 1) * numsec + (j + 1) % numsec];
				}
			}

			for (int j = 0; j < numsec; j++)
			{
				FaceList.Add(new Face());
				((Face)FaceList[FaceList.Count - 1]).vList[0] = (Vertex)VertexList[(j + 1) % numsec];
				((Face)FaceList[FaceList.Count - 1]).vList[1] = (Vertex)VertexList[j];
				((Face)FaceList[FaceList.Count - 1]).vList[2] = (Vertex)VertexList[VertexList.Count - 2];
			}

			for (int j = 0; j < numsec; j++)
			{
				FaceList.Add(new Face());
				((Face)FaceList[FaceList.Count - 1]).vList[0] = (Vertex)VertexList[VertexList.Count - 1];
				((Face)FaceList[FaceList.Count - 1]).vList[1] = (Vertex)VertexList[numrow * numsec + j];
				((Face)FaceList[FaceList.Count - 1]).vList[2] = (Vertex)VertexList[numrow * numsec + (j + 1) % numsec];
			}

			vList = new Vertex[VertexList.Count];
			fList = new Face[FaceList.Count];

			for (int i = 0; i < VertexList.Count; i++) vList[i] = (Vertex)VertexList[i];
			for (int i = 0; i < FaceList.Count; i++) fList[i] = (Face)FaceList[i];

			GetXYSideKeyPoints();

		}
		public Cuboid(byte level)
		{
			ArrayList VertexList = new ArrayList();
			ArrayList FaceList = new ArrayList();

			numrow = MyElement.myRect.level; // divid the z-axis
			numsec = 4 * level; // divid the circles

		//	numrow = level; // divid the z-axis
		//	numsec = 4 * level; // divid the circles

			double lenrow = 1.0 / numrow;
			double z_bottom = -0.5;

			int indexer = 0;
			int ncircles = numrow + 1; // number of circles

			
			double width = MyElement.myRect.width;
		/*	Vector2d[] x = new Vector2d[4] {
				new Vector2d(-width, -width),
				new Vector2d(width, -width),
				new Vector2d(width, width),
				new Vector2d(-width, width)
			};
			double[] r = new double[level];
			for (int i = 0; i < level; ++i)
			{
				r[i] = (double)i / level;
			}
			Vector2d[][] p = new Vector2d[4][];
			for (int i = 0; i < 4; ++i)
			{
				p[i] = new Vector2d[level];
			}
			for (int i = 0; i < numsec; i++)
			{
				int k = i / level;
				int l = i % level;
				int t = (k + 1) % 4;
				// compute x,y coords
				p[k][l] = x[k] * (1 - r[l]) + x[t] * r[l];
			}
			*/

			for (int i = 0; i < ncircles; i++)
			{
				for (int j = 0; j < numsec; j++)
				{
					int k = j / level;
					int l = j % level;

					double x = MyElement.myRect.points[j].x;
					double y = MyElement.myRect.points[j].y;

					VertexList.Add(
						new Vertex(
							x, 
							y, 
							z_bottom + i * lenrow
						)
					);
					indexer++;
				}
				// key points
				if (i == 0 || i == numrow)
				{
					int id = indexer - numsec;
					for (int j = 0; j < numsec; ++j)
					{
						this.zSideKeyPoints.Add(id++);
					}
				}
			}

			// bottom and top plane ...
			double dy = 2.0 * width / level;
			for (int i = 0; i < 2; ++i)
			{
				double zz = i == 0 ? -0.5 : 0.5;
				for (int j = 1; j < level; ++j)
				{
					for (int k = 1; k < level; ++k)
					{
						double x = MyElement.myRect.points[k].x;
						double y = MyElement.myRect.points[k].y;

						Vector2d q = new Vector2d(x, y + dy * j);
						VertexList.Add(
							new Vertex(q.x,q.y,zz)
						);
					}
				}
			}

			for (int i = 0; i < VertexList.Count; i++)
			{
				((Vertex)VertexList[i]).Index = i;
			}

			for (int i = 0; i < numrow; i++)
			{
				for (int j = 0; j < numsec; j++)
				{
					FaceList.Add(new Face());
					FaceList.Add(new Face());

					((Face)FaceList[2 * (i * numsec + j)]).vList[0] = (Vertex)VertexList[i * numsec + j];
					((Face)FaceList[2 * (i * numsec + j)]).vList[1] = (Vertex)VertexList[(i + 1) * numsec + (j + 1) % numsec];
					((Face)FaceList[2 * (i * numsec + j)]).vList[2] = (Vertex)VertexList[(i + 1) * numsec + j];

					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[0] = (Vertex)VertexList[i * numsec + j];
					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[1] = (Vertex)VertexList[i * numsec + (j + 1) % numsec];
					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[2] = (Vertex)VertexList[(i + 1) * numsec + (j + 1) % numsec];
				}
			}

			// triangles in top and bottom plane
			int[][] line = new int[level + 1][];
			for (int jj = 0; jj < 2; ++jj)
			{
				int start = jj == 0 ? 0 : numrow * numsec;
				for (int i = 0; i < level + 1; ++i)
				{
					line[i] = new int[level + 1];
					if (i == 0)
					{
						for (int j = 0; j < level + 1; ++j)
						{
							line[i][j] = j + start;
						}
					}
					else if (i == level)
					{
						for (int count = 0, j = level * 3; j >= level * 2; --j, ++count)
						{
							line[i][count] = j + start;
						}
					}
					else
					{
						int s = numsec - i;
						int t = level + i;
						int w = jj == 0 ? 0 : (level-1)*(level-1);
						int v = numsec * ncircles + w;
						line[i][0] = s + start;
						line[i][level] = t + start;
						for (int j = 0; j < level-1; ++j)
						{
							line[i][j + 1] = v + (i - 1) * (level - 1) + j;
						}
					}
				}
				for (int i = 0; i < level; ++i)
				{
					for (int j = 0; j < level; ++j)
					{
						int s = line[i][j], t = line[i][j + 1];
						int h = line[i + 1][j], q = line[i + 1][j + 1];
						
						Face f1 = new Face(); Face f2 = new Face();
						if (jj == 1)
						{
							f1.vList[0] = (Vertex)VertexList[s];
							f1.vList[1] = (Vertex)VertexList[t];
							f1.vList[2] = (Vertex)VertexList[q];

							f2.vList[0] = (Vertex)VertexList[s];
							f2.vList[1] = (Vertex)VertexList[q];
							f2.vList[2] = (Vertex)VertexList[h];
						}
						else
						{
							f1.vList[0] = (Vertex)VertexList[s];
							f1.vList[1] = (Vertex)VertexList[q];
							f1.vList[2] = (Vertex)VertexList[t];

							f2.vList[0] = (Vertex)VertexList[s];
							f2.vList[1] = (Vertex)VertexList[h];
							f2.vList[2] = (Vertex)VertexList[q];
						}

						FaceList.Add(f1);
						FaceList.Add(f2);
					}
				}
			}


			vList = new Vertex[VertexList.Count];
			fList = new Face[FaceList.Count];

			for (int i = 0; i < VertexList.Count; i++) vList[i] = (Vertex)VertexList[i];
			for (int i = 0; i < FaceList.Count; i++) fList[i] = (Face)FaceList[i];

			GetXYSideKeyPoints();

		}
		public void GetXYSideKeyPoints()
		{
			List<int>[] lines = new List<int>[12];
			int m = numsec / 4;
			int n = numrow + 1;
			for (int i = 0; i < 12; ++i)
			{
				lines[i] = new List<int>();
				if (i < 4 || i > 7)
				{
					int s = i < 4 ? i * m : numrow * numsec + (i - 8) * m;
					for (int j = 0; j <= m; ++j)
					{
						int k = s + j;
						if ((i == 3 && j == m) ||
							(i == 11 && j == m))
						{
							k-=numsec;
						}
						lines[i].Add(k);
					}
				}
				else
				{
					int s = (i - 4) * m;
					for (int j = 0; j < n; ++j)
					{
						lines[i].Add(s + j*numsec);
					}
				}
			}

			// assign --
			int[][] permutation = new int[4][]
			{
				new int[4]{1,6,9,5}, /// x side 
				new int[4]{3,4,11,7},/// x side
				new int[4]{5,8,4,0}, /// y side
				new int[4]{7,10,6,2} /// y side
			};
			
			List<int> list = null;
			this.xSideKeyPoints.Clear(); this.ySideKeyPoints.Clear();
			for (int i = 0; i < 4; ++i)
			{
				list = i < 2 ? this.xSideKeyPoints : this.ySideKeyPoints;

				int l0 = permutation[i][0];
				int l1 = permutation[i][1];
				int l2 = permutation[i][2];
				int l3 = permutation[i][3];

				if (i < 2)
				{
					for (int j = 0; j < lines[l0].Count; ++j)			///+
					{
						list.Add(lines[l0][j]);
					}
					for (int j = 1; j < lines[l1].Count; ++j)			///+
					{
						list.Add(lines[l1][j]);
					}
					for (int j = lines[l2].Count - 2; j >= 0; --j)		///-
					{
						list.Add(lines[l2][j]);
					}
					for (int j = lines[l3].Count - 2; j > 0; --j)		///-
					{
						list.Add(lines[l3][j]);
					}
				}
				else
				{
					for (int j = 0; j < lines[l0].Count; ++j)			///+
					{
						list.Add(lines[l0][j]);
					}
					for (int j = lines[l1].Count - 2; j >= 0; --j)		///-
					{
						list.Add(lines[l1][j]);
					}
					for (int j = lines[l2].Count - 2; j >= 0; --j)		///-
					{
						list.Add(lines[l2][j]);
					}
					for (int j = 1; j < lines[l3].Count - 1; ++j)		///+
					{
						list.Add(lines[l3][j]);
					}
				}
			}
		}
		public Cuboid()
		{
			vList = new Vertex[8];
			fList = new Face[12];

			double z_bottom = -0.5, z_top = 0.5;

			vList[0] = new Vertex(1, 1, z_bottom);
			vList[1] = new Vertex(-1, 1, z_bottom);
			vList[2] = new Vertex(-1, -1, z_bottom);
			vList[3] = new Vertex(1, -1, z_bottom);
			vList[4] = new Vertex(1, 1, z_top);
			vList[5] = new Vertex(-1, 1, z_top);
			vList[6] = new Vertex(-1, -1, z_top);
			vList[7] = new Vertex(1, -1, z_top);

			for (int i = 0; i < vList.Count(); i++)
			{
				vList[i].Index = i;
				keyPoints.Add(i); // key points
			}

			fList[0] = new Face();
			fList[0].vList[0] = vList[0];
			fList[0].vList[1] = vList[1];
			fList[0].vList[2] = vList[5];

			fList[1] = new Face();
			fList[1].vList[0] = vList[0];
			fList[1].vList[1] = vList[5];
			fList[1].vList[2] = vList[4];

			fList[2] = new Face();
			fList[2].vList[0] = vList[1];
			fList[2].vList[1] = vList[2];
			fList[2].vList[2] = vList[6];

			fList[3] = new Face();
			fList[3].vList[0] = vList[1];
			fList[3].vList[1] = vList[6];
			fList[3].vList[2] = vList[5];

			fList[4] = new Face();
			fList[4].vList[0] = vList[2];
			fList[4].vList[1] = vList[3];
			fList[4].vList[2] = vList[7];

			fList[5] = new Face();
			fList[5].vList[0] = vList[2];
			fList[5].vList[1] = vList[7];
			fList[5].vList[2] = vList[6];

			fList[6] = new Face();
			fList[6].vList[0] = vList[3];
			fList[6].vList[1] = vList[0];
			fList[6].vList[2] = vList[4];

			fList[7] = new Face();
			fList[7].vList[0] = vList[3];
			fList[7].vList[1] = vList[4];
			fList[7].vList[2] = vList[7];

			fList[8] = new Face();
			fList[8].vList[0] = vList[3];
			fList[8].vList[1] = vList[1];
			fList[8].vList[2] = vList[0];

			fList[9] = new Face();
			fList[9].vList[0] = vList[3];
			fList[9].vList[1] = vList[2];
			fList[9].vList[2] = vList[1];

			fList[10] = new Face();
			fList[10].vList[0] = vList[4];
			fList[10].vList[1] = vList[5];
			fList[10].vList[2] = vList[7];

			fList[11] = new Face();
			fList[11].vList[0] = vList[5];
			fList[11].vList[1] = vList[6];
			fList[11].vList[2] = vList[7];
		}
	}
	public class Cone : Polyhedron
	{
		public Cone(int level)
		{
			ArrayList VertexList = new ArrayList();
			ArrayList FaceList = new ArrayList();

			int numrow = 3 * level;
			int numsec = 4 * level;

			double lenrow = 1.0 / numrow;
			double radius = 0.5;
			double angle = 2 * Math.PI / numsec;

			for (int j = 0; j < numsec; j++)
				VertexList.Add(new Vertex(0.0 * radius * Math.Cos(j * angle), 0.0 * radius * Math.Sin(j * angle), 0));

			for (int i = 0; i < numrow; i++)
			{
				for (int j = 0; j < numsec; j++)
					VertexList.Add(new Vertex((((double)i + 1.0) / (double)numrow) * radius * Math.Cos(j * angle), (((double)i + 1.0) / (double)numrow) * radius * Math.Sin(j * angle), (i + 1) * lenrow));
			}

			for (int i = 0; i < VertexList.Count; i++)
			{
				((Vertex)VertexList[i]).Index = i;
			}

			for (int i = 0; i < numrow; i++)
			{
				for (int j = 0; j < numsec; j++)
				{
					FaceList.Add(new Face());
					FaceList.Add(new Face());

					((Face)FaceList[2 * (i * numsec + j)]).vList[0] = (Vertex)VertexList[i * numsec + j];
					((Face)FaceList[2 * (i * numsec + j)]).vList[1] = (Vertex)VertexList[(i + 1) * numsec + (j + 1) % numsec];
					((Face)FaceList[2 * (i * numsec + j)]).vList[2] = (Vertex)VertexList[(i + 1) * numsec + j];

					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[0] = (Vertex)VertexList[i * numsec + j];
					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[1] = (Vertex)VertexList[i * numsec + (j + 1) % numsec];
					((Face)FaceList[2 * (i * numsec + j) + 1]).vList[2] = (Vertex)VertexList[(i + 1) * numsec + (j + 1) % numsec];
				}
			}

			vList = new Vertex[VertexList.Count];
			fList = new Face[FaceList.Count];

			for (int i = 0; i < VertexList.Count; i++) vList[i] = (Vertex)VertexList[i];
			for (int i = 0; i < FaceList.Count; i++) fList[i] = (Face)FaceList[i];
		}
	}
	public class Sphere : Polyhedron
	{
		public int level = 8;
		public double radius = 1.0;
		public Sphere(byte level)
		{
			ArrayList VertexList = new ArrayList();
			ArrayList FaceList = new ArrayList();

			double t = 2 * Math.PI / level;
			double s = 2 * Math.PI / level;
			int rows = level / 2;
			for (int i = 0; i <= rows; ++i)
			{
				if (i == 0 || i == rows)
				{
					double x = 0;
					double y = 0;
					double z = i == 0 ? -radius : radius;
					VertexList.Add(
						new Vertex(x, y, z)
					);
				}
				else
				{
					double phi = -Math.PI / 2 + i * s;
					double r = radius * Math.Cos(phi);
					double z = radius * Math.Sin(phi);
					for (int j = 0; j < level; ++j)
					{
						double theta = j * t;
						double x = r * Math.Cos(theta);
						double y = r * Math.Sin(theta);
						VertexList.Add(
							new Vertex(x, y, z)
						);
					}
				}
			}

			for (int i = 0; i < VertexList.Count; i++)
			{
				((Vertex)VertexList[i]).Index = i;
			}

			for (int i = 0; i < rows; ++i)
			{
				if (i == 0)
				{
					int s0 = 0;
					for (int j = 0; j < level; ++j)
					{
						int s1 = j + 1;
						int s2 = (j + 1) % level + 1;

						Face f1 = new Face();

						f1.vList[0] = (Vertex)VertexList[s0];
						f1.vList[1] = (Vertex)VertexList[s2];
						f1.vList[2] = (Vertex)VertexList[s1];

						FaceList.Add(f1);
					}
				}
				else if (i == rows - 1)
				{
					int s0 = VertexList.Count - 1;
					for (int j = 0; j < level; ++j)
					{
						int ss = (i-1) * level;
						int s1 = j + ss + 1;
						int s2 = (j + 1) % level + ss + 1;

						Face f1 = new Face();

						f1.vList[0] = (Vertex)VertexList[s1];
						f1.vList[1] = (Vertex)VertexList[s2];
						f1.vList[2] = (Vertex)VertexList[s0];

						FaceList.Add(f1);
					}
				}
				else
				{
					for (int j = 0; j < level; ++j)
					{
						int ss = (i-1) * level;
						int s0 = j + ss + 1;
						int s1 = (j+1) % level + ss + 1;
						int s2 = s0 + level;
						int s3 = s1 + level;

						Face f1 = new Face();
						Face f2 = new Face();
						
						f1.vList[0] = (Vertex)VertexList[s0];
						f1.vList[1] = (Vertex)VertexList[s3];
						f1.vList[2] = (Vertex)VertexList[s2];

						f2.vList[0] = (Vertex)VertexList[s0];
						f2.vList[1] = (Vertex)VertexList[s1];
						f2.vList[2] = (Vertex)VertexList[s3];

						FaceList.Add(f1);
						FaceList.Add(f2);
					}
				}
			}

			// key points
			int lv = rows / 2; int shift = (lv-1)*level+1;
			for (int i = 0; i < level; ++i)
			{
				this.keyPoints.Add(i+shift);
			}
			
			/*
			// xz cross-section
			this.keyPoints.Add(0);
			for (int i = 1; i < rows; ++i)
			{
				this.keyPoints.Add(1+(i-1)*level);
			}
			int last = VertexList.Count - 1, start = last - rows;
			this.keyPoints.Add(last);
			for (int i = 1; i < rows; ++i)
			{
				this.keyPoints.Add(start - (i - 1) * level);
			}

			// yz cross-section
			this.keyPoints.Add(0); start = 1 + level / 4;
			for (int i = 1; i < rows; ++i)
			{
				this.keyPoints.Add(start + (i - 1) * level);
			}
			start = last - level / 4;
			this.keyPoints.Add(last);
			for (int i = 1; i < rows; ++i)
			{
				this.keyPoints.Add(start - (i - 1) * level);
			}
			*/

			vList = new Vertex[VertexList.Count];
			fList = new Face[FaceList.Count];

			for (int i = 0; i < VertexList.Count; i++) vList[i] = (Vertex)VertexList[i];
			for (int i = 0; i < FaceList.Count; i++) fList[i] = (Face)FaceList[i];
		}

	}
}
