﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MyGeometry;


namespace ShapeGraph
{	
	/// util classes 
	// edge class, connect two points
	public class Edge
	{
		public int i = -1;		// start vertex
		public int j = -1;		// end vertex
		public int prevI = -1;  // prev end vertex
		public int prevJ = -1;  // prev end vertex
		public double coordx = 0; // for reconstruction
		public double coordy = 0;
		public double coordz = 0;
		public double cost = 0; // edge cost
		public int host = -1;
		public Edge(int s, int t)
		{
			prevI = i = s;
			prevJ = j = t;
		}
	}
	// local frame class, for vertex
	public class LocalFrame : ICloneable
	{
		public Vector3d b1;
		public Vector3d b2;
		public Vector3d N;
		public Vector3d old_b1 = new Vector3d();	// original value
		public Vector3d old_b2 = new Vector3d();
		public Vector3d old_N = new Vector3d();
		#region clonable interface
		public object Clone()
		{
			LocalFrame obj = new LocalFrame();

			obj.b1 = new Vector3d(this.b1);
			obj.b2 = new Vector3d(this.b2);
			obj.N = new Vector3d(this.N);
			obj.Copy();

			return obj;
		}
		#endregion
		public void Copy()
		{
			old_b1.x = b1.x;
			old_b1.y = b1.y;
			old_b1.z = b1.z;

			old_b2.x = b2.x;
			old_b2.y = b2.y;
			old_b2.z = b2.z;

			old_N.x = N.x;
			old_N.y = N.y;
			old_N.z = N.z;
		}
		public void Normalize()
		{
			this.b1 = b1.Normalize();
			this.b2 = b2.Normalize();
			this.N = N.Normalize();
		}
	}
	public class KeyPoint /// point controller
	{
		public int index = -1;
		public Proxy hostProxy = null;
		public FeatureCurve encodeCurve = null; /// encoder curve, basically is the nearest one
		public Proxy encodeProxy = null;
		public int encodePointIndex = -1; /// encoded keypoint index
		public Vector3d pos;									  
		public double a = 0;
		public double b = 0;
		public double r = 0;
		public KeyPoint(Proxy p, int i)
		{
			this.hostProxy = p;
			this.index = i;
		}
	}
	public class ControlCircle
	{
		public List<int> vertexIndexes = null;
		public Vector3d oldCenter;
		public Vector3d center;
		public double scale = 1;
		public Vector3d xAxis;
		public Vector3d yAxis;
		public Vector3d zAxis;
		public ControlCircle(List<int> ptIndexes, Vector3d c, double s, Vector3d x, Vector3d y, Vector3d z)
		{
			this.vertexIndexes = ptIndexes;
			this.center = c;
			this.oldCenter = new Vector3d(c);
			this.scale = s;
			this.xAxis = x;
			this.yAxis = y;
			this.zAxis = z;
		}
		public ControlCircle()
		{ }
	}
	public class FeatureCurve /// feature curve controller, for each proxy
	{
		/// to remmeber which curve on earth it is!!!
		public static int currCurveIndex = 0;
		public int globalId = -1;

		public List<KeyPoint> keyPoints = new List<KeyPoint>();
		public List<FeatureCurve> orthogonalPartners = null;
		public List<FeatureCurve> reflectivePartners = new List<FeatureCurve>();
		public List<FeatureCurve> rotationalPartners = new List<FeatureCurve>();
		public Proxy hostProxy = null;
		public bool hasCoplarnarRelations = false; /// if so, we retain its center during optimization
		public bool isFixed = false;
		public int inProxyIndex = -1;
		public int propagatorId = -1;

		/// whether its three axis are fixed, 
		/// if so, we do not change the scale along the fixed axis
		public bool isFixingXScale = false;
		public bool isFixingYScale = false;
		public bool isFixingZScale = false;

		public bool positionUpdated = false;

		public List<CurveReflectiveSymmetry> reflectiveSymmetries = new List<CurveReflectiveSymmetry>();
		public List<CurveRotationalSymmetry> rotationalSymmetries = new List<CurveRotationalSymmetry>();
		public int concentricGroupId = -1;
		public int coplanarGroupId = -1;
		public int parallelGroupId = -1;
		public int similarityGroupId = -1;

		public static bool needPartialEnlarge = false;

		public bool propagated = false; /// being propagated or not
		public bool visited = false;
		public bool bridged = false;
		public Vector3d centroid;
		public Vector3d oldCentroid;
		public Vector3d scale;
		public Vector3d oldScale;
		public Vector3d initialScale;
		public double radius = -1;
		public int propagateIndex = -1;
		public int index = -1;
		public int oldId = -1;
		public LocalFrame frame = null;
		public LocalFrame[] controlFrame = null;
		public LocalFrame[] encodingFrame = null;
		public List<Edge> bridges = null;
		public void ComputeEncodeFrames(bool firsttime)
		{
			int n = this.keyPoints.Count;

			if (this.encodingFrame == null)
			{
				this.encodingFrame = new LocalFrame[n];
				for (int i = 0; i < n; ++i)
				{
					this.encodingFrame[i] = new LocalFrame();
				}
			}

			for (int i = 0; i < n; ++i)
			{
				int s = (n + i - 1) % n, t = (i + 1) % n;

				Vector3d v0 = this.keyPoints[s].pos;
				Vector3d v1 = this.keyPoints[i].pos;
				Vector3d v2 = this.keyPoints[t].pos;

				Vector3d cx = (v2 - v1).Normalize();
				Vector3d cz = this.frame.N.Normalize();
				Vector3d cy = cz.Cross(cx).Normalize();

				if (double.IsNaN(cx.x) || double.IsNaN(cy.x) || double.IsNaN(cz.x))
				{
					Program.OutputText("Error in updating encoding frame! curve: " + this.index.ToString(),true);
				}
				else
				{
					this.encodingFrame[i].b1 = cx;
					this.encodingFrame[i].b2 = cy;
					this.encodingFrame[i].N = cz;
				}

				if (firsttime)
				{
					this.encodingFrame[i].Copy();
				}
			}
		}
		public void ComputeControlFrame()
		{
			if (FeatureCurve.needPartialEnlarge && this.hostProxy.type == Proxy.ProxyType.Cuboid)
			{
				ComputeControlFrameBySegments();
				return;
			}

			int n = this.keyPoints.Count;

			if (this.controlFrame == null) this.controlFrame = new LocalFrame[n];
			
			for (int i = 0; i < n; ++i)
			{
				this.controlFrame[i] = new LocalFrame();

				this.controlFrame[i].b1 = new Vector3d(this.frame.b1);
				this.controlFrame[i].b2 = new Vector3d(this.frame.b2);
				this.controlFrame[i].N = new Vector3d(this.frame.N);
				this.controlFrame[i].Copy();
			}
		}
		public FeatureCurve(Proxy host, int id)
		{
			this.hostProxy = host;
			this.inProxyIndex = id;
		}
		public void DiliverSymmetryHandles()
		{
			Proxy p = this.hostProxy;
			foreach (CurveReflectiveSymmetry symm in this.reflectiveSymmetries)
			{
				FeatureCurve c = symm.member1 == this ? symm.member2 : symm.member1;
				if (c.hostProxy != this.hostProxy)
					c.hostProxy.SetHandleCurve(c);
			}
			foreach (CurveRotationalSymmetry symm in this.rotationalSymmetries)
			{
				foreach (FeatureCurve c in symm.members)
				{
					if (c != this)
					{
						c.hostProxy.AllowAxisEnlarge = true;
						c.hostProxy.SetHandleCurve(c);
					}
				}
			}
		}
		public void GetSymmetryPartners()
		{
			this.reflectivePartners.Clear();
			this.rotationalPartners.Clear();
			foreach (CurveReflectiveSymmetry symm in this.reflectiveSymmetries)
			{
				FeatureCurve c = symm.member1 == this ? symm.member2 : symm.member1;
				this.reflectivePartners.Add(c);
			}
			foreach (CurveRotationalSymmetry symm in this.rotationalSymmetries)
			{
				foreach (FeatureCurve c in symm.members)
				{
					if (c != this)
					{
						this.rotationalPartners.Add(c);
					}
				}
			}
		}
		public void UpdateControlFrame()
		{
			if (FeatureCurve.needPartialEnlarge && this.hostProxy.type == Proxy.ProxyType.Cuboid)
			{
				ComputeControlFrameBySegments();
				UpdateControlFrameBySegments();
				return;
			}

			Vector3d s1 = this.scale, s2 = this.initialScale;
			double x = s1.x / s2.x;
			double y = s1.y / s2.y;
			double z = s1.z / s2.z;

			if (this.hostProxy.type == Proxy.ProxyType.Sphere)
			{
				y = z = x;
			}

			double xRisidual = Math.Abs(1 - x);
			double yRisidual = Math.Abs(1 - y);
			double zRisidual = Math.Abs(1 - z);
			double err = 1e-3;
			Matrix4d S = Matrix4d.IdentityMatrix();
			Matrix4d Q = Matrix4d.IdentityMatrix();
			
			Vector3d b1 = this.frame.b1.Normalize();
			Vector3d b2 = this.frame.b2.Normalize();
			Vector3d N = this.frame.N.Normalize();

			b2 = N.Cross(b1).Normalize();
			b1 = b2.Cross(N).Normalize();

			if (xRisidual > err || yRisidual > err || zRisidual > err)
			{
				S = Matrix4d.ScalingMatrix(x, y, z);

				Matrix4d R = new Matrix4d(
					new Matrix3d(
						b1, b2, N
					)
				);
				R[3, 3] = 1;
				R = R.Transpose();

				Q = R.Inverse() * S * R;
			}

			LocalFrame frm = new LocalFrame();
			frm.b1 = (Q * new Vector4d(b1, 0)).XYZ();
			frm.b2 = (Q * new Vector4d(b2, 0)).XYZ();
			frm.N = (Q * new Vector4d(N, 0)).XYZ();

			for (int i = 0; i < this.controlFrame.Length; ++i)
			{
				this.controlFrame[i] = frm.Clone() as LocalFrame;
			}
		}

		public void ComputeControlFrameBySegments()
		{
			int n = this.keyPoints.Count;

			this.controlFrame = new LocalFrame[n];

			int c0 = 0, c1 = n / 4, c2 = n / 2, c3 = n * 3 / 4;
			double[] pos = this.hostProxy.FitMesh.VertexPos;

			Vector3d u0 = new Vector3d(pos, this.keyPoints[c0].index * 3);
			Vector3d u1 = new Vector3d(pos, this.keyPoints[c1].index * 3);
			Vector3d u2 = new Vector3d(pos, this.keyPoints[c2].index * 3);
			Vector3d u3 = new Vector3d(pos, this.keyPoints[c3].index * 3);

			Vector3d e0 = (u1 - u0).Normalize();
			Vector3d e1 = (u2 - u1).Normalize();
			Vector3d e2 = (u3 - u2).Normalize();
			Vector3d e3 = (u0 - u3).Normalize();

			Vector3d z = this.frame.N.Normalize();
			for (int i = 0; i < c1; ++i)
			{
				LocalFrame frm = new LocalFrame();
				frm.b1 = e0;
				frm.N = z;
				frm.b2 = z.Cross(e0).Normalize();
				frm.Copy();

				this.controlFrame[i] = frm;
			}
			for (int i = c1; i < c2; ++i)
			{
				LocalFrame frm = new LocalFrame();
				frm.b1 = e1;
				frm.N = z;
				frm.b2 = z.Cross(e1).Normalize();
				frm.Copy();

				this.controlFrame[i] = frm;
			}
			for (int i = c2; i < c3; ++i)
			{
				LocalFrame frm = new LocalFrame();
				frm.b1 = e2;
				frm.N = z;
				frm.b2 = z.Cross(e2).Normalize();
				frm.Copy();

				this.controlFrame[i] = frm;
			}
			for (int i = c3; i < n; ++i)
			{
				LocalFrame frm = new LocalFrame();
				frm.b1 = e3;
				frm.N = z;
				frm.b2 = z.Cross(e3).Normalize();
				frm.Copy();

				this.controlFrame[i] = frm;
			}
		}
		public void UpdateControlFrameBySegments()
		{
			int n = this.keyPoints.Count;

			int[] c = new int[4] {0, n / 4, n / 2, n * 3 / 4};

			double[] opos = this.hostProxy.fittedPosition;
			double[] pos = this.hostProxy.FitMesh.VertexPos;

			int a0 = this.keyPoints[c[0]].index * 3;
			int a1 = this.keyPoints[c[1]].index * 3;
			int a2 = this.keyPoints[c[2]].index * 3;
			int a3 = this.keyPoints[c[3]].index * 3;

			Vector3d u0 = new Vector3d(pos, a0);
			Vector3d u1 = new Vector3d(pos, a1);
			Vector3d u2 = new Vector3d(pos, a2);
			Vector3d u3 = new Vector3d(pos, a3);

			Vector3d v0 = new Vector3d(opos, a0);
			Vector3d v1 = new Vector3d(opos, a1);
			Vector3d v2 = new Vector3d(opos, a2);
			Vector3d v3 = new Vector3d(opos, a3);

			double s0 = (u1 - u0).Length() / (v1 - v0).Length();
			double s1 = (u2 - u1).Length() / (v2 - v1).Length();
			double s2 = (u3 - u2).Length() / (v3 - v2).Length();
			double s3 = (u0 - u3).Length() / (v0 - v3).Length();
			double[] s = new double[4] { s0, s1, s2, s3 };

			double zSclae = this.scale.z / this.initialScale.z;
			double thr = 1e-3;
			for (int i = 0; i < 4; ++i)
			{
				if (Math.Abs(1 - s[i]) < thr)
					s[i] = 1;
			}
			if (Math.Abs(1 - zSclae) < thr) zSclae = 1;


			FeatureCurve opposite = this.hostProxy.GetOppositeCurve(this);

			for (int i = 0; i < 4; ++i)
			{
				int b0 = c[i], b1 = c[(i + 1) % 4];

				if (i == 3) b1 = n;

				Vector3d z = this.frame.N.Normalize();
				Vector3d x = this.controlFrame[b0].b1;
				Vector3d y = this.controlFrame[b0].b2;

				Matrix4d R = new Matrix4d(
					new Matrix3d(
						x, y, z
					)
				);
				R[3, 3] = 1;
				R = R.Transpose();

				Matrix4d S = Matrix4d.ScalingMatrix(s[i], 1, zSclae);

				int bt = GetOppositePointIndex(b0);
				double osy = (new Vector3d(opos, this.keyPoints[b0].index * 3) - 
					new Vector3d(opos, this.keyPoints[bt].index * 3)).Length();
				double osz = (new Vector3d(opos, this.keyPoints[b0].index * 3) -
					new Vector3d(opos, opposite.keyPoints[b0].index * 3)).Length();

				Matrix4d Q = R.Inverse() * S * R;
				for (int j = b0; j < b1; ++j)
				{
					int jt = GetOppositePointIndex(j);
					double sy = (new Vector3d(pos, this.keyPoints[jt].index * 3) - 
						new Vector3d(pos, this.keyPoints[j].index * 3)).Length();
					double yScale = sy / osy;

					double sz = (new Vector3d(pos, this.keyPoints[j].index * 3) -
						new Vector3d(pos, opposite.keyPoints[j].index * 3)).Length();
					double zzScale = sz / osz;


					bool needUpdate = false;
					if (Math.Abs(1 - yScale) > thr)
					{
						needUpdate = true;
						S[1,1] = yScale;
					}
					if (Math.Abs(1 - zzScale) > thr)
					{
						needUpdate = true;
						S[2,2] = zzScale;
					}

					if (needUpdate)
					{
						Q = R.Inverse() * S * R;
					}

					LocalFrame frm = this.controlFrame[j];

					this.controlFrame[j].b1 = (Q * new Vector4d(frm.b1, 0)).XYZ();
					this.controlFrame[j].b2 = (Q * new Vector4d(frm.b2, 0)).XYZ();
					this.controlFrame[j].N = (Q * new Vector4d(frm.N, 0)).XYZ();
				}
			}
		}
		public void UpdateControlFrameBySegments_old()
		{
			int n = this.keyPoints.Count;

			int[] c = new int[4] { 0, n / 4, n / 2, n * 3 / 4 };

			double[] opos = this.hostProxy.fittedPosition;
			double[] pos = this.hostProxy.FitMesh.VertexPos;

			int a0 = this.keyPoints[c[0]].index * 3;
			int a1 = this.keyPoints[c[1]].index * 3;
			int a2 = this.keyPoints[c[2]].index * 3;
			int a3 = this.keyPoints[c[3]].index * 3;

			Vector3d u0 = new Vector3d(pos, a0);
			Vector3d u1 = new Vector3d(pos, a1);
			Vector3d u2 = new Vector3d(pos, a2);
			Vector3d u3 = new Vector3d(pos, a3);

			Vector3d v0 = new Vector3d(opos, a0);
			Vector3d v1 = new Vector3d(opos, a1);
			Vector3d v2 = new Vector3d(opos, a2);
			Vector3d v3 = new Vector3d(opos, a3);

			double s0 = (u1 - u0).Length() / (v1 - v0).Length();
			double s1 = (u2 - u1).Length() / (v2 - v1).Length();
			double s2 = (u3 - u2).Length() / (v3 - v2).Length();
			double s3 = (u0 - u3).Length() / (v0 - v3).Length();
			double[] s = new double[4] { s0, s1, s2, s3 };

			double zSclae = this.scale.z / this.initialScale.z;
			double thr = 1e-3;
			for (int i = 0; i < 4; ++i)
			{
				if (Math.Abs(1 - s[i]) < thr)
					s[i] = 1;
			}
			if (Math.Abs(1 - zSclae) < thr) zSclae = 1;

			double scalX = this.scale.x / this.initialScale.x;
			double scalY = this.scale.y / this.initialScale.y;
			double scalZ = this.scale.z / this.initialScale.z;

			for (int i = 0; i < 4; ++i)
			{
				int b0 = c[i], b1 = c[(i + 1) % 4];

				if (i == 3) b1 = n;

				Vector3d z = this.frame.N.Normalize();
				Vector3d x = this.controlFrame[b0].b1;
				Vector3d y = this.controlFrame[b0].b2;

				Matrix4d R = new Matrix4d(
					new Matrix3d(
						x, y, z
					)
				);
				R[3, 3] = 1;
				R = R.Transpose();

				Matrix4d S = Matrix4d.ScalingMatrix(scalX, scalY, scalZ);

				Matrix4d Q = R.Inverse() * S * R;
				for (int j = b0; j < b1; ++j)
				{
					LocalFrame frm = this.controlFrame[j];

					this.controlFrame[j].b1 = (Q * new Vector4d(frm.b1, 0)).XYZ();
					this.controlFrame[j].b2 = (Q * new Vector4d(frm.b2, 0)).XYZ();
					this.controlFrame[j].N = (Q * new Vector4d(frm.N, 0)).XYZ();
				}
			}
		}
		private int GetOppositePointIndex(int i)
		{
			int n = this.keyPoints.Count;
			int m = n / 4;
			int k = i/m;
			if (k % 2 == 0)
			{
				return m * 3 - i;
			}
			else
			{
				return (m * 5 - i) % n;
			}
		}

		public void UpdateKeyPointsPos()
		{
			double[] pos = this.hostProxy.FitMesh.VertexPos;
			foreach (KeyPoint kpt in this.keyPoints)
			{
				kpt.pos = new Vector3d(pos, kpt.index * 3);
			}
		}
		public void UpdateMeshPos()
		{
			double[] pos = this.hostProxy.FitMesh.VertexPos;
			foreach (KeyPoint kpt in this.keyPoints)
			{
				int b = kpt.index * 3;
				pos[b] = kpt.pos.x;
				pos[b+1] = kpt.pos.y;
				pos[b+2] = kpt.pos.z;
			}
		}

		public Vector3d GetCentroid()
		{
			Vector3d center = new Vector3d();

			int n = this.keyPoints.Count;
			if (this.hostProxy.type == Proxy.ProxyType.Cuboid)
			{
				int m = n / 4;
				for (int i = 0, j = 0; i < 4; ++i, j += m)
				{
					center += this.keyPoints[j].pos;
				}
				center /= 4;
			}
			else
			{
				for (int i = 0; i < n; ++i)
				{
					center += this.keyPoints[i].pos;
				}
				center /= n;
			}
			return center;
		}
		public Vector3d GetXYZScale(Vector3d center)
		{
			int npoints = this.keyPoints.Count;
			int m = npoints / 4;

			double yscale = 0, xscale = 0;

			for (int i = 0; i < m; ++i)
			{
				Vector3d v0 = this.keyPoints[i].pos;
				Vector3d v1 = this.keyPoints[i + 1].pos;

				double dist = HelperFunction.Dist2Line(center, v0, v1);

				yscale += dist;
			}
			for (int i = m * 2; i < m * 3; ++i)
			{
				Vector3d v0 = this.keyPoints[i].pos;
				Vector3d v1 = this.keyPoints[i + 1].pos;

				double dist = HelperFunction.Dist2Line(center, v0, v1);

				yscale += dist;
			}
			yscale /= (npoints / 2);

			for (int i = m; i < m * 2; ++i)
			{
				Vector3d v0 = this.keyPoints[i].pos;
				Vector3d v1 = this.keyPoints[i + 1].pos;

				double dist = HelperFunction.Dist2Line(center, v0, v1);

				xscale += dist;
			}
			for (int i = m * 3; i < npoints; ++i)
			{
				Vector3d v0 = this.keyPoints[i].pos;
				Vector3d v1 = this.keyPoints[(i + 1) % npoints].pos;

				double dist = HelperFunction.Dist2Line(center, v0, v1);

				xscale += dist;
			}
			xscale /= (npoints / 2);

			Vector3d s = new Vector3d(xscale, yscale, this.scale.z);

			return s;
		}
	}
	public class FeatureCurvePair : PriorityQueueElement
	{
		public FeatureCurve p = null;
		public FeatureCurve q = null;
		public double distance = 0;
		public FeatureCurvePair(FeatureCurve i, FeatureCurve j, double dis)
		{
			this.p = i;
			this.q = j;
			this.distance = dis;
		}
		#region PriorityQueueElement Members
		private int pqIndex = -1;
		public int PQIndex
		{
			get
			{
				return pqIndex;
			}
			set
			{
				pqIndex = value;
			}
		}
		public int CompareTo(PriorityQueueElement obj)
		{
			FeatureCurvePair r = obj as FeatureCurvePair;
			if (this.distance < r.distance) return -1;
			if (this.distance > r.distance) return 1;
			return 0;
		}

		#endregion
	}
	public class FeatureCurveGroup
	{
		public List<FeatureCurve> members = new List<FeatureCurve>();
		public GroupStyle style = GroupStyle.None;
		public int id = -1;
		public bool propagated = false;
		public bool optimized = false;
		public Vector3d planeNormal;
		public FeatureCurveGroup(List<FeatureCurve> g, int _id)
		{
			this.members = g;
			this.id = _id;
		}
	}

	public class CurveReflectiveSymmetry
	{
		public FeatureCurve member1 = null;
		public FeatureCurve member2 = null;
		public Vector3d planeCenter;
		public Vector3d planeNormal;
		public CurveReflectiveSymmetry()
		{
		}
		public CurveReflectiveSymmetry(FeatureCurve i, FeatureCurve j)
		{
			this.member1 = i;
			this.member2 = j;
		}
		public void SetMirror(Vector3d c, Vector3d n)
		{
			this.planeCenter = c;
			this.planeNormal = n;
		}
	}
	public class CurveRotationalSymmetry
	{
		public List<FeatureCurve> members = null;
		public Vector3d rotCenter;
		public Vector3d rotAxis;
		public double rotAngle = 0;
		public CurveRotationalSymmetry()
		{
		}
		public CurveRotationalSymmetry(List<FeatureCurve> membrs)
		{
			this.members = membrs;
		}
		public void SetRotInfos(Vector3d c, Vector3d axis, double angle)
		{
			this.rotCenter = c;
			this.rotAxis = axis;
			this.rotAngle = angle;
		}
	}
	public class ProxyReflectiveSymmetry
	{
		public Proxy member1 = null;
		public Proxy member2 = null;
		public Vector3d planeCenter;
		public Vector3d planeNormal;
		public ProxyReflectiveSymmetry()
		{
		}
		public ProxyReflectiveSymmetry(Proxy i, Proxy j)
		{
			this.member1 = i;
			this.member2 = j;
		}
		public void SetMirror(Vector3d c, Vector3d n)
		{
			this.planeCenter = c;
			this.planeNormal = n;
		}
	}
	public class ProxyRotationalSymmetry
	{
		public List<Proxy> members = null;
		public Vector3d rotCenter;
		public Vector3d rotAxis;
		public double rotAngle = 0;
		public ProxyRotationalSymmetry()
		{
		}
		public ProxyRotationalSymmetry(List<Proxy> membrs)
		{
			this.members = membrs;
		}
		public void SetRotInfos(Vector3d c, Vector3d axis, double angle)
		{
			this.rotCenter = c;
			this.rotAxis = axis;
			this.rotAngle = angle;
		}
	}

	public class UserCurve
	{
		public int numofPoints = 0;
		public double depth = 0;
		public List<Vector3d> samplePoints = null;
		public List<Vector3d> pointsDerivatives = null;
		public UserCurve(List<Vector3d> samples, List<Vector3d> derivs)
		{
			this.samplePoints = samples;
			this.pointsDerivatives = derivs;
			this.numofPoints = samples.Count;
		}
		public UserCurve(Vector3d[] samples, Vector3d[] derivs)
		{
			this.numofPoints = samples.Length;
			this.samplePoints = new List<Vector3d>();
			this.pointsDerivatives = new List<Vector3d>();
			for (int i = 0; i < numofPoints; ++i)
			{
				this.samplePoints.Add(samples[i]);
				this.pointsDerivatives.Add(derivs[i]);
			}
		}
	}

	public class Proxy
	{
		public enum ProxyType
		{
			Sphere = 0,
			Cylinder,
			Cuboid,
			Cone,
			TruncatedCylinder,
			TruncatedCuboid,
			TruncatedCone,
			GeneralizedCylinder,
			None,
		}
		public ProxyType type = ProxyType.Sphere;
		public static ProxyType WhatProxy(int id)
		{
			ProxyType t = ProxyType.None;
			
			switch (id)
			{
				case 0: 
					t = ProxyType.Sphere; 
					break;
				case 1: 
					t = ProxyType.Cylinder; 
					break;
				case 2: 
					t = ProxyType.Cuboid; 
					break;
				case 3: 
					t = ProxyType.Cone; 
					break;
				case 4: 
					t = ProxyType.TruncatedCylinder; 
					break;
				case 5: 
					t = ProxyType.TruncatedCuboid; 
					break;
				case 6: 
					t = ProxyType.TruncatedCone; 
					break;
				case 7:
					t = ProxyType.GeneralizedCylinder;
					break;
			}

			return t;
		}

		private class VertexRecord : PriorityQueueElement
		{
			private int pqIndex = -1;
			private int index = -1;
			private double distance = double.MaxValue;

			public VertexRecord(int index)
			{
				this.index = index;
			}
			public int Index { get { return index; } }
			public double Distance
			{
				get { return distance; }
				set { distance = value; }
			}

			#region PriorityQueueElement Members

			public int PQIndex
			{
				get
				{
					return pqIndex;
				}
				set
				{
					pqIndex = value;
				}
			}

			#endregion
			#region IComparable<PriorityQueueElement> Members

			public int CompareTo(PriorityQueueElement obj)
			{
				VertexRecord rec = obj as VertexRecord;
				if (this.distance < rec.distance) return -1;
				if (this.distance > rec.distance) return 1;
				return 0;
			}

			#endregion
		}

		public Proxy()
		{
			FitScale = new Vector3d(1.0, 1.0, 1.0);
			FitRotateAxis = new Vector3d(0.0, 0.0, 1.0);
			FitRotateAngle = 0.0;
			FitTranslate = new Vector3d(0.0, 0.0, 0.0);

			OrigMesh = null;
			FitMesh = null;
		}

		public bool IsTopLevelProxy = false;
		public bool IsSelfReflective = false;
		public Proxy parent = null;
		public List<Proxy> children = null; /// children proxyies, for hierarchical propagation
		public int colorIndex = -1;

		public List<Proxy> insideProxies = null; /// children proxyies, for uer defined proxies


		public Matrix4d FittingMatrix()
		{
			Matrix4d translation = Matrix4d.TranslationMatrix(FitTranslate);
			Matrix4d rotation = Matrix4d.RotationMatrix(FitRotateAxis, FitRotateAngle);
			Matrix4d scaling = Matrix4d.ScalingMatrix(FitScale.x, FitScale.y, FitScale.z);

			Matrix4d transform = translation * rotation * scaling;

			return transform;
		}


		public List<Proxy> closestNeighbors = null;


		public bool onedegreeFreedom = false;
		public bool AllowAxisEnlarge = false;
		public bool reflected = false;
		public void Update(
			Vector3d x, Vector3d y, Vector3d z,
			Vector3d center,
			Vector3d scale
		)
		{
			// check whethe it is inside out
			Vector3d zz = x.Cross(y).Normalize();
			if (zz.Dot(z) < 0)
			{
				this.reflected = true;
			}
			else
			{
				this.reflected = false;
			}
			z = zz;

			switch (this.type)
			{
				case ProxyType.Sphere:
					{
						double d = scale.x > scale.y ? (scale.x > scale.z ? scale.x : scale.z) : 
							(scale.y > scale.z ? scale.y : scale.z);
						scale.z = scale.y = scale.x = d;
					}
					break;
				case ProxyType.Cylinder:
					{
						double d = scale.x > scale.y ? scale.x : scale.y;
						scale.y = scale.x = d;
						scale.z *=2;
					}
					break;
				case ProxyType.Cuboid:
					scale.z *= 2;
					break;
			}

			// translation, rotation and scaling matrixes
			Matrix4d T = Matrix4d.TranslationMatrix(center);
			Matrix4d S = Matrix4d.ScalingMatrix(scale.x, scale.y, scale.z);
			Matrix4d R = new Matrix4d(new Matrix3d(x, y, z)); // rotation
			R[3, 3] = 1;

			Matrix4d Q = T * R * S;

			this.PrinAxis1 = x;
			this.PrinAxis2 = y;
			this.Axis = z;

		//	this.x = new Vector3d(x);
		//	this.y = new Vector3d(y);
		//	this.z = new Vector3d(axis);
		//	this.o = new Vector3d(center);

			this.FitRotateAxis = x;
			this.FitRotateAngle = 0;
			this.FitTranslate = center;

			this.UpdateFitMesh(
				Q, 
				this.startPosition
			);
		}
		public bool transcated = false;
		public bool rescaled = false;
		public bool isFixingXScale = false;
		public bool isFixingYScale = false;
		public bool isFixingZScale = false;
		public void Update
		(
			Vector3d x, Vector3d y, Vector3d z, // x, y, z axis
			Vector3d center, 
			Vector3d scale,
			double topXEnlarge, double topYEnlarge,
			double botXEnlarge, double botYEnlarge, 
			double zEnlarge,
			int whichAxis
		)
		{

			if (this.isFixingXScale) scale.x = this.FitScale.x;
			if (this.isFixingYScale) scale.y = this.FitScale.y;
			if (this.isFixingZScale) scale.z = this.FitScale.z;

			this.FitScale = scale;

			if (!AllowAxisEnlarge && !transcated)
			{
				Update(x, y, z, center, scale);
				return;
			}

			if (this.type == ProxyType.Cuboid)
			{
				this.FitTranslate = center;
				((CuboidProxy)this).UpdateCurveOrientation(whichAxis);
				((CuboidProxy)this).UpdateShape(whichAxis);
				return;
			}

			// check whethe it is inside out
			Vector3d zz = x.Cross(y).Normalize();
			if (zz.Dot(z) < 0)
			{
				this.reflected = true;
			}
			else
			{
				this.reflected = false;
			}
			z = zz;

			Matrix4d S = Matrix4d.IdentityMatrix();
			S[3, 3] = 1;
			// first update static positions
			switch (this.type)
			{
				case ProxyType.Cylinder:
					this.UpdateInitPosition_Cylinder(
						topXEnlarge, 
						botXEnlarge, 
						zEnlarge
					);
					this.transcated = true;
					break;
				case ProxyType.Sphere:
					S[0, 0] = S[1, 1] = S[2, 2] = scale.z;
					break;
			}

			// translation, rotation and scaling matrixes
			Matrix4d T = Matrix4d.TranslationMatrix(center);
			Matrix4d R = new Matrix4d(new Matrix3d(x, y, z)); // rotation
			R[3, 3] = 1;

			Matrix4d Q = T * R * S;

		//	this.PrinAxis1 = x;
		//	this.PrinAxis2 = y;
		//	this.Axis = axis;

		//	this.x = new Vector3d(x);
		//	this.y = new Vector3d(y);
		//	this.z = new Vector3d(axis);
		//	this.o = new Vector3d(center);

			this.FitRotateAxis = x;
			this.FitRotateAngle = 0;
			this.FitTranslate = center;

			this.UpdateFitMesh(Q, this.initialPosition);
		}
		public void UpdateFitMesh(Matrix4d T, double[] position)
		{
			for (int i = 0, j = 0; i < OrigMesh.VertexCount; i++, j += 3)
			{
				Vector3d v = new Vector3d(position, j);
				v = (T * new Vector4d(v, 1)).XYZ();

				FitMesh.VertexPos[j] = v.x;
				FitMesh.VertexPos[j + 1] = v.y;
				FitMesh.VertexPos[j + 2] = v.z;
			}
			
			{
				this.AssignKeyPointsPos();
				
				this.ResetLocalFrames();
				this.UpdateEncodingFrame();
			}
		}
		public void ResetLocalFrames()
		{
			FitMesh.ComputeNormals();
			FitMesh.InitLocalFrames();
		}
		public void ScaleLocalFrame(Matrix4d S)
		{
			Matrix4d T = Matrix4d.TranslationMatrix(this.FitTranslate);
		
			Matrix4d R = new Matrix4d(
				new Matrix3d(
					this.PrinAxis1, 
					this.PrinAxis2, 
					this.Axis
				)
			); // rotation
			R[3, 3] = 1;
			R = R.Transpose();

			Matrix4d Q = R.Inverse() * S * R;
		 /** */

			LocalFrame[] frames = this.FitMesh.local_frames;
			for (int i = 0; i < frames.Length; ++i)
			{
				Vector3d b1 = frames[i].b1;
				Vector3d b2 = frames[i].b2;
				Vector3d N = frames[i].N;

				b1 = (Q * new Vector4d(b1, 0)).XYZ();
				b2 = (Q * new Vector4d(b2, 0)).XYZ();
				N = (Q * new Vector4d(N, 0)).XYZ();

				frames[i].b1 = b1;
				frames[i].b2 = b2;
				frames[i].N = N;
			}

			this.rescaled = true;
		}
		public void DeformUpdate()
		{
			this.FitMesh.ComputeNormals();

			Proxy.UpdateOrigMesh(this);

			foreach (FeatureCurve c in this.featuredCurves)
			{
				foreach (LocalFrame frm in c.encodingFrame)
					frm.Copy();
				c.frame.Copy();
				c.oldScale = new Vector3d(c.scale);
			}

			this.GetFeatureCurveScales();
			this.ComputeFeatureCurvesCentroid();
			this.ComputeFeatureCurvesRadius();

			if (this.type == ProxyType.GeneralizedCylinder)
			{
				GeneralizedCylinderProxy g = (GeneralizedCylinderProxy)this;
				g.UpdateOldCenters();
				g.UpdateCenter();
			}

			/// coordinates
			this.o = new Vector3d(this.FitTranslate);
			this.x = new Vector3d(this.PrinAxis1);
			this.y = new Vector3d(this.PrinAxis2);
			this.z = new Vector3d(this.Axis);
			this.oldFitScale = new Vector3d(this.FitScale);
		}


		public void AssignFixScaleAxis(int whichAxis, bool fix)
		{
			if (this.type == ProxyType.Cuboid)
			{
				List<FeatureCurve> lst = this.featuredCurves;
				switch (whichAxis)
				{
					case 0:
						{
							lst[0].isFixingXScale = fix;
							lst[1].isFixingXScale = fix;
							lst[2].isFixingZScale = fix;
							lst[3].isFixingZScale = fix;
							lst[4].isFixingYScale = fix;
							lst[5].isFixingYScale = fix;
						}
						break;
					case 1:
						{
							lst[0].isFixingYScale = fix;
							lst[1].isFixingYScale = fix;
							lst[2].isFixingXScale = fix;
							lst[3].isFixingXScale = fix;
							lst[4].isFixingZScale = fix;
							lst[5].isFixingZScale = fix;
						}
						break;
					case 2:
						{
							lst[0].isFixingZScale = fix;
							lst[1].isFixingZScale = fix;
							lst[2].isFixingYScale = fix;
							lst[3].isFixingYScale = fix;
							lst[4].isFixingXScale = fix;
							lst[5].isFixingXScale = fix;
						}
						break;
				}
			}
			else
			{
				foreach (FeatureCurve c in this.featuredCurves)
				{
					switch (whichAxis)
					{
						case 0:
							c.isFixingXScale = fix;
							break;
						case 1:
							c.isFixingYScale = fix;
							break;
						case 2:
							c.isFixingZScale = fix;
							break;
					}
				}
			}
		}

		public void CheckTranscatedState()
		{
			Proxy p = this;

			double thr = 0.1;
			if (p.transcated && p.handleCurve == null)
			{
				if (p.type == Proxy.ProxyType.Cylinder || p.onedegreeFreedom)
				{
					Vector3d s1 = p.featuredCurves[0].scale;
					Vector3d s2 = p.featuredCurves[1].scale;
					double sx = Math.Abs(s1.x - s2.x) / (s1.x + s2.x);
					double sy = Math.Abs(s1.y - s2.y) / (s1.y + s2.y);
					double sz = Math.Abs(s1.z - s2.z) / (s1.z + s2.z);
					if (sx < thr || sy < thr || sz < thr)
					{
						p.transcated = false;
						p.AllowAxisEnlarge = false;
					}
				}
				else
				{
					for (int i = 0; i < 6; i += 2)
					{
						Vector3d s1 = p.featuredCurves[i].scale;
						Vector3d s2 = p.featuredCurves[i + 1].scale;
						double sx = Math.Abs(s1.x - s2.x) / (s1.x + s2.x);
						double sy = Math.Abs(s1.y - s2.y) / (s1.y + s2.y);
						double sz = Math.Abs(s1.z - s2.z) / (s1.z + s2.z);
						if (sx < thr || sy < thr || sz < thr)
						{
							p.transcated = false;
							p.AllowAxisEnlarge = false;
							break;
						}
					}
				}
			}
		}

		public FeatureCurve GetOppositeCurve(FeatureCurve c)
		{
			int id = this.featuredCurves.IndexOf(c);
			if (id == -1) return null;

			int j = 1 - id;
			if (this.type == ProxyType.Cuboid && !this.onedegreeFreedom)
			{
				j = id <= 1 ? 1 - id : id <= 3 ? 5 - id : 9 - id;
			}
			return this.featuredCurves[j];
		}

		public void ComputeCurveInfos()
		{
			this.GetFeatureCurveScales();
			this.ComputeFeatureCurvesCentroid();
			this.ComputeFeatureCurvesFrame();
			this.ComputeFeatureCurvesRadius();
		}

		// update original mesh
		public static void UpdateOrigMesh(Proxy proxy)
		{
			// copy to original mesh, for deformation
			Mesh msh = proxy.OrigMesh;
			msh.VertexPos = proxy.FitMesh.VertexPos.Clone() as double[];
			msh.VertexNormal = proxy.FitMesh.VertexNormal.Clone() as double[];
			msh.FaceNormal = proxy.FitMesh.FaceNormal.Clone() as double[];
		//	proxy.OrigMesh.ComputeNormals();

			LocalFrame[] lcf1 = proxy.FitMesh.local_frames;
			if (msh.local_frames == null)
			{
				msh.local_frames = new LocalFrame[lcf1.Length];
				for (int i = 0; i < lcf1.Length; ++i)
					msh.local_frames[i] = new LocalFrame();
			}
			LocalFrame[] lcf2 = msh.local_frames;
			for (int i = 0; i < lcf1.Length; ++i)
			{
				LocalFrame lc = lcf1[i];
				lcf2[i].b1 = new Vector3d(lc.b1);
				lcf2[i].b2 = new Vector3d(lc.b2);
				lcf2[i].N = new Vector3d(lc.N);
				lcf2[i].Copy();
			}
		}

		// create proxy
		public static Proxy MakeProxy(ProxyType type, Vector3d pos, Vector3d scale, Vector3d x, Vector3d y, Vector3d z)
		{
			Proxy proxy = Proxy.Create(type);
			if (proxy != null)
			{
				proxy.type = type;
				proxy.FitScale = scale;
				proxy.oldFitScale = new Vector3d(scale);
				
				proxy.Update(x, y, z, pos, scale);
				proxy.AssignKeyPointsPos();
				proxy.ComputeFeatureCurvesCentroid();
				proxy.ComputeFeatureCurvesRadius();
				proxy.ComputeFeatureCurvesFrame();

				UpdateOrigMesh(proxy);

				proxy.x = proxy.PrinAxis1; // copy
				proxy.y = proxy.PrinAxis2;
				proxy.z = proxy.Axis;
				proxy.o = proxy.FitTranslate;
				proxy.fittedPosition = proxy.FitMesh.VertexPos.Clone() as double[];
			}
			
			return proxy;
		}
		public static Proxy Create(ProxyType type)
		{
			Proxy proxy = null;
			switch (type)
			{
				case ProxyType.Sphere: proxy = new SphereProxy(); break;
				case ProxyType.Cylinder: proxy = new CylinderProxy(); break;
				case ProxyType.Cuboid: proxy = new CuboidProxy(); break;
				case ProxyType.Cone: break;
				case ProxyType.TruncatedCylinder: break;
				case ProxyType.TruncatedCuboid: break;
				case ProxyType.TruncatedCone: break;
			}
			return proxy;
		}

		private bool curveInitilScaleComputed = false;
		public void GetFeatureCurveScales()
		{
			Vector3d s = this.FitScale;
			if (this.transcated)
			{
				if (this.type == ProxyType.Cylinder)
				{
					foreach (FeatureCurve c in this.featuredCurves)
					{
						c.oldScale.z = c.scale.z = s.z;
					}
				}
				else if (this.type == ProxyType.Cuboid)
				{
					foreach (FeatureCurve c in this.featuredCurves)
						c.scale = c.GetXYZScale(c.centroid);

					if (!this.onedegreeFreedom)
					{
						this.featuredCurves[2].oldScale.z = this.featuredCurves[2].scale.z = s.x;
						this.featuredCurves[3].oldScale.z = this.featuredCurves[3].scale.z = s.x;
						this.featuredCurves[4].oldScale.z = this.featuredCurves[4].scale.z = s.y;
						this.featuredCurves[5].oldScale.z = this.featuredCurves[5].scale.z = s.y;
					}
					this.featuredCurves[0].oldScale.z = this.featuredCurves[0].scale.z = s.z;
					this.featuredCurves[1].oldScale.z = this.featuredCurves[1].scale.z = s.z;

				/*	int axis = ((CuboidProxy)this).whichAxisTranscated;
					switch (axis)
					{
						case 0:
							this.featuredCurves[2].oldScale.z = this.featuredCurves[2].scale.z = s.x;
							this.featuredCurves[3].oldScale.z = this.featuredCurves[3].scale.z = s.x;
							break;
						case 1:
							this.featuredCurves[4].oldScale.z = this.featuredCurves[4].scale.z = s.y;
							this.featuredCurves[5].oldScale.z = this.featuredCurves[5].scale.z = s.y;
							break;
						case 2:
							this.featuredCurves[0].oldScale.z = this.featuredCurves[0].scale.z = s.z;
							this.featuredCurves[1].oldScale.z = this.featuredCurves[1].scale.z = s.z;
							break;
					}*/
				}
				return;
			}
			
			if (this.type == ProxyType.Cuboid && !onedegreeFreedom)
			{
				this.featuredCurves[0].scale = new Vector3d(s);
				this.featuredCurves[1].scale = new Vector3d(s);
				this.featuredCurves[0].oldScale = new Vector3d(s);
				this.featuredCurves[1].oldScale = new Vector3d(s);

				this.featuredCurves[2].scale = new Vector3d(s.y, s.z, s.x);
				this.featuredCurves[3].scale = new Vector3d(s.y, s.z, s.x);
				this.featuredCurves[2].oldScale = new Vector3d(s.y, s.z, s.x);
				this.featuredCurves[3].oldScale = new Vector3d(s.y, s.z, s.x);

				this.featuredCurves[4].scale = new Vector3d(s.z, s.x, s.y);
				this.featuredCurves[5].scale = new Vector3d(s.z, s.x, s.y);
				this.featuredCurves[4].oldScale = new Vector3d(s.z, s.x, s.y);
				this.featuredCurves[5].oldScale = new Vector3d(s.z, s.x, s.y);
			}
			else
			{
				foreach (FeatureCurve c in this.featuredCurves)
				{
					c.scale = new Vector3d(s);
					c.oldScale = new Vector3d(s);
				}
			}

			if (!curveInitilScaleComputed)
			{
				curveInitilScaleComputed = true;
				foreach (FeatureCurve c in this.featuredCurves)
				{
					c.initialScale = new Vector3d(c.scale);
				}
			}
		}
		public void ComputeFeatureCurvesFrame()
		{
			switch (this.type)
			{
				case ProxyType.Sphere:
					{
						FeatureCurve c1 = this.featuredCurves[0];
						c1.frame = new LocalFrame();

						c1.frame.b1 = new Vector3d(this.PrinAxis1);
						c1.frame.b2 = new Vector3d(this.PrinAxis2);
						c1.frame.N = new Vector3d(this.Axis);

						c1.frame.Copy();
					}
					break;
				case ProxyType.Cylinder:
				case ProxyType.Cuboid:
					{

						FeatureCurve c1 = this.featuredCurves[0];
						FeatureCurve c2 = this.featuredCurves[1];
						c1.frame = new LocalFrame();
						c2.frame = new LocalFrame();

						c1.frame.b1 = new Vector3d() - this.PrinAxis1;
						c1.frame.b2 = new Vector3d(this.PrinAxis2);
						c1.frame.N = new Vector3d() - this.Axis;

						c2.frame.b1 = new Vector3d(this.PrinAxis1);
						c2.frame.b2 = new Vector3d(this.PrinAxis2);
						c2.frame.N = new Vector3d(this.Axis);

						c1.frame.Copy();
						c2.frame.Copy();

						if (type == ProxyType.Cuboid && !onedegreeFreedom)
						{
							FeatureCurve c3 = this.featuredCurves[3];
							FeatureCurve c4 = this.featuredCurves[2];

							c3.frame = new LocalFrame();
							c4.frame = new LocalFrame();

							c3.frame.b1 = new Vector3d() - this.PrinAxis2;
							c3.frame.b2 = new Vector3d(this.Axis);
							c3.frame.N = new Vector3d() - this.PrinAxis1;

							c4.frame.b1 = new Vector3d(this.PrinAxis2);
							c4.frame.b2 = new Vector3d(this.Axis);
							c4.frame.N = new Vector3d(this.PrinAxis1);

							c3.frame.Copy();
							c4.frame.Copy();

							FeatureCurve c5 = this.featuredCurves[4];
							FeatureCurve c6 = this.featuredCurves[5];

							c5.frame = new LocalFrame();
							c6.frame = new LocalFrame();

							c5.frame.b1 = new Vector3d() - this.Axis;
							c5.frame.b2 = new Vector3d(this.PrinAxis1);
							c5.frame.N = new Vector3d() - this.PrinAxis2;

							c6.frame.b1 = new Vector3d(this.Axis);
							c6.frame.b2 = new Vector3d(this.PrinAxis1);
							c6.frame.N = new Vector3d(this.PrinAxis2);

							c5.frame.Copy();
							c6.frame.Copy();
						}

						if (type == ProxyType.Cuboid && !onedegreeFreedom && this.transcated)
						{
							double[] pos = this.FitMesh.VertexPos;
							foreach (FeatureCurve c in this.featuredCurves)
							{
								int nn = c.keyPoints.Count;
								Vector3d u = c.keyPoints[0].pos;
								Vector3d v = c.keyPoints[nn / 4].pos;
								Vector3d w = c.keyPoints[nn / 2].pos;

								Vector3d n = (u - v).Cross(w - v).Normalize();

								Vector3d N = c.frame.N.Normalize();
								double cos = N.Dot(n);
								if (cos < 0)
								{
									cos = -cos;
									n = new Vector3d() - n;
								}

								if (cos > HelperFunction.cos3) continue;

								Vector3d rotAxis = N.Cross(n).Normalize();

								if (double.IsNaN(rotAxis.x)) continue;

								double rotAngle = Math.Acos(cos);

								Matrix4d R = Matrix4d.RotationMatrix(rotAxis, rotAngle);

								Vector3d x = (R * new Vector4d(c.frame.b1, 0)).XYZ();
								Vector3d y = (R * new Vector4d(c.frame.b2, 0)).XYZ();

								c.frame.b1 = x;
								c.frame.b2 = y;
								c.frame.N = n;

								c.frame.Copy();
							}
						}
					}
					break;
				case ProxyType.GeneralizedCylinder:
					{
						FeatureCurve c1 = this.featuredCurves[0];
						FeatureCurve c2 = this.featuredCurves[1];
						c1.frame = new LocalFrame();
						c2.frame = new LocalFrame();

						int n = c1.keyPoints.Count;
						int i = c1.keyPoints[0].index;
						int j = c1.keyPoints[n/4].index;

						double[] pos = c1.hostProxy.FitMesh.VertexPos;
						Vector3d vi = new Vector3d(pos, i * 3) - c1.centroid;
						Vector3d vj = new Vector3d(pos, j * 3) - c1.centroid;

						Vector3d n1 = vj.Cross(vi).Normalize();
						Vector3d x1 = vi.Normalize();
						Vector3d y1 = n1.Cross(x1).Normalize();

						c1.frame.b1 = x1;
						c1.frame.b2 = y1;
						c1.frame.N = n1;


						i = c2.keyPoints[0].index;
						j = c2.keyPoints[n / 4].index;

						Vector3d ui = new Vector3d(pos, i * 3) - c2.centroid;
						Vector3d uj = new Vector3d(pos, j * 3) - c2.centroid;

						Vector3d n2 = ui.Cross(uj).Normalize();
						Vector3d x2 = ui.Normalize();
						Vector3d y2 = n2.Cross(x2).Normalize();

						c2.frame.b1 = x2;
						c2.frame.b2 = y2;
						c2.frame.N = n2;

						c1.frame.Copy();
						c2.frame.Copy();
					}
					break;
			}
		}
		public void ComputeFeatureCurvesCentroid()
		{
			foreach (FeatureCurve c in this.featuredCurves)
			{
				Vector3d center = new Vector3d();
				foreach (KeyPoint pt in c.keyPoints)
				{
					Vector3d p = pt.pos;
					center += p;
				}
				c.centroid = center / c.keyPoints.Count;
				c.oldCentroid = new Vector3d(c.centroid);
			}
		}
		public void ComputeFeatureCurvesRadius()
		{
			foreach (FeatureCurve c in this.featuredCurves)
			{
				double sum = 0;
				foreach (KeyPoint pt in c.keyPoints)
				{
					Vector3d p = pt.pos;
					sum += (p-c.centroid).Length();
				}
				c.radius = sum / c.keyPoints.Count;
			}
		}
		public void ComputeFeatureCurveEncodeFrames(bool firsttime)
		{
			foreach (FeatureCurve c in this.featuredCurves)
			{
				c.ComputeEncodeFrames(firsttime);
				if (firsttime)
				{
					c.ComputeControlFrame();
				}
			}	
		}

		public void AssignKeyPointsPos()
		{
			foreach (FeatureCurve c in this.featuredCurves)
			{
				c.UpdateKeyPointsPos();
			}
		}

		// for propagation
		public bool tagged = false; // whether used or not.
		public bool IsFixed = false; // do not move in propagation
		public bool Refitted = false;
		public bool editted = false; // if edited by user, need update the bridges
		public bool optimized = false;
		public bool propagated = false;
		public int similarityGroupId = -1;


		public List<byte> groupId = new List<byte>(); // which group
		public int propagatorId = -1;
		public Matrix4d propagateMatrix = null;
		public Matrix4d storedTransformation = null;
		public List<ProxyRotationalSymmetry> rotationalSymmetries = new List<ProxyRotationalSymmetry>();
		public List<ProxyReflectiveSymmetry> reflectiveSymmetries = new List<ProxyReflectiveSymmetry>();
		public int oldId = -1;
		public int id = -1;
		public List<FeatureCurve> featuredCurves = null;
		public List<FeatureCurve> childrenHandleCurves = null;

		public double fittingError = -1;
		public double[] startPosition = null; // position when created.
		public double[] initialPosition = null; // position when created.
		public double[] fittedPosition = null; // position when first time fitted.

		public LocalFrame[] encodingFrame = null;
		public void UpdateEncodingFrame()
		{
			int n = this.FitMesh.VertexCount;

			bool initialization = false;
			if (this.encodingFrame == null)
			{
				initialization = true;
				this.encodingFrame = new LocalFrame[n];
				for (int i = 0; i < n; ++i)
					this.encodingFrame[i] = new LocalFrame();
			}

			Mesh m = this.FitMesh;
			for (int i = 0, j = 0; i < n; ++i, j += 3)
			{
				int adj = this.FitMesh.AdjVV[i][0];
				Vector3d N = new Vector3d(m.VertexNormal, j);
				Vector3d p = new Vector3d(m.VertexPos, j);
				Vector3d q = new Vector3d(m.VertexPos, adj * 3);
				Vector3d e = q - p;
				Vector3d b = (e - e.Dot(N) * N).Normalize();
				
				encodingFrame[i].b1 = b;
				encodingFrame[i].N = N.Normalize();
				encodingFrame[i].b2 = N.Cross(b).Normalize();
				
				if (initialization)
					encodingFrame[i].Copy();
			}
		}

		public List<int> handlePoints = new List<int>();
		public List<int> associatePts = null; // corresponding pts to the original mesh, should be assigned to on fitting.
		public List<int> connectedNbrs = new List<int>(); // connected proxies
		
		public Vector3d handleCenter = new Vector3d();
		public Vector3d oldHandleCenter = new Vector3d();
		public FeatureCurve handleCurve = null;
		public List<FeatureCurve> controlCurves = null;
		public void SetHandleCurve(FeatureCurve curv)
		{
			if (this.type == ProxyType.Sphere ||
				id < 0) return; // no key points handle for sphere

			this.AllowAxisEnlarge = true;
			this.handleCurve = curv;

			this.handlePoints.Clear();
			foreach (KeyPoint pt in curv.keyPoints)
				FitMesh.Flag[pt.index] = 0;

			SetKeyPtAsHandle();
			FindHandleCenter();
		}
		public void SetKeyPtAsHandle()
		{
			if (this.handleCurve == null) return;
			ClearHandlePoints();
			foreach (KeyPoint kpt in this.handleCurve.keyPoints)
			{
				int vIdx = kpt.index;
				this.handlePoints.Add(vIdx);
				FitMesh.Flag[vIdx] = 1;
			}
		}
		public void ClearHandlePoints()
		{
			this.handlePoints.Clear();
			foreach (FeatureCurve curve in this.featuredCurves)
			{
				foreach (KeyPoint kpt in curve.keyPoints)
				{
					FitMesh.Flag[kpt.index] = 0;
				}
			}
		}
		public void CopyHandlePoints(Proxy p)
		{
			Vector3d n = (this.o - p.o).Normalize();
			Vector3d z = HelperFunction.ReflectionVector(p.z, n);

			if (p.handleCurve == null) 
				this.handleCurve = null;
			else if (z.Dot(this.z) < 0) // reflected
			{
				this.handleCurve = this.featuredCurves[1 - p.handleCurve.inProxyIndex];
			}
			else
				this.handleCurve = this.featuredCurves[p.handleCurve.inProxyIndex];

			SetKeyPtAsHandle();
			foreach (int i in handlePoints)
				FitMesh.Flag[i] = 1;
			FindHandleCenter();
		}
		private void FindHandleCenter()
		{
			if (this.handlePoints.Count == 0) return;

			// find center
			double[] pos = FitMesh.VertexPos;
			Vector3d c = new Vector3d();
			foreach (int v in this.handlePoints)
			{
				Vector3d p = new Vector3d(pos, v * 3);
				c += p;
			}
			c /= handlePoints.Count;
			this.handleCenter = c;
			this.oldHandleCenter = new Vector3d(c);
		}
		public void SymmetryCommunicate()
		{
			if (this.reflectiveSymmetries != null)
				foreach (ProxyReflectiveSymmetry symm in this.reflectiveSymmetries)
				{
					Proxy opp = symm.member1 == this ? symm.member2 : symm.member1;

					opp.AllowAxisEnlarge = this.AllowAxisEnlarge;
					opp.CopyHandlePoints(this);
				}
			if (this.rotationalSymmetries != null)
				foreach (ProxyRotationalSymmetry symm in this.rotationalSymmetries)
				{
					foreach (Proxy opp in symm.members)
					{
						if (opp != this)
						{
							opp.AllowAxisEnlarge = this.AllowAxisEnlarge;
							opp.CopyHandlePoints(this);
						}
					}
				}
		}

		// get rotation and translation matrix from axis and center translation
		public Matrix4d GetRotationMatrix()
		{
			double cos = this.z.Dot(this.Axis);
			if (cos > Math.Cos(5.0 / 180))
				return Matrix4d.IdentityMatrix();
			Vector3d rot_axis = this.z.Cross(this.Axis).Normalize();
			double rot_angle = Math.Acos(cos);
			Matrix4d R = Matrix4d.RotationMatrix(rot_axis, rot_angle);
			return R;
		}
		public Matrix4d GetTranslationMatrix()
		{
			return Matrix4d.TranslationMatrix(this.FitTranslate - this.o);
		}

		// local axises
		public Vector3d FitTranslate;
		public Vector3d PrinAxis1;
		public Vector3d PrinAxis2;
		public Vector3d Axis;

		public Vector3d o; // original 
		public Vector3d x;
		public Vector3d y;
		public Vector3d z;

		public byte currRotationAxis = 2;

		// Fit Transformation
		public Vector3d oldFitScale;
		public Vector3d FitScale;
		public Vector3d FitRotateAxis;
		public double FitRotateAngle;
		
		// Fit Enlarge

		public double xAxisYTopEnlarge = 1;
		public double xAxisYBotEnlarge = 1;
		public double xAxisZTopEnlarge = 1;
		public double xAxisZBotEnlarge = 1;
		public double xAxisEnlarge = 1;

		public double yAxisXTopEnlarge = 1;
		public double yAxisXBotEnlarge = 1;
		public double yAxisZTopEnlarge = 1;
		public double yAxisZBotEnlarge = 1;
		public double yAxisEnlarge = 1;

		public double zAxisXTopEnlarge = 1;
		public double zAxisXBotEnlarge = 1;
		public double zAxisYTopEnlarge = 1;
		public double zAxisYBotEnlarge = 1;
		public double zAxisEnlarge = 1;


		public double topXEnlarge = 1;
		public double topYEnlarge = 1;
		public double botXEnlarge = 1;
		public double botYEnlarge = 1;
		public double zEnlarge = 1; // principle axis


		// For Truncated cylinder/cubiod Only
		public int numrow = 0; // # of circles
		public int numsec = 0; // # of sections per circle
		private void UpdateInitPosition_Cylinder(double topEnlarge, double botEnlarge, double zLength)
		{
			double[] pos = this.initialPosition;

			double z_bottom = -zLength/2;
			double z_top = zLength / 2;

			double lenrow = zLength / numrow;
			double radius = 1;
			double angle = 2 * Math.PI / numsec;

			int indexer = 0;
			int ncircles = numrow + 1;
			for (int i = 0; i < ncircles; i++)
			{
				// positions
				radius = topEnlarge * ((ncircles - i) / (double)ncircles) + 
					(i/(double)ncircles)*botEnlarge;
				for (int j = 0; j < numsec; j++)
				{
					int b = indexer * 3;
					double x = radius * Math.Cos(j * angle);
					double y = radius * Math.Sin(j * angle);
					double z = z_bottom + i * lenrow;
					pos[b] = x;
					pos[b + 1] = y;
					pos[b + 2] = z;
					indexer++;
				}
			}
			int c = indexer * 3;
			pos[c + 2] = z_bottom; // bottom vertex
			pos[c + 5] = z_top; // top vertex
		}
		private void  UpdateInitPosition_Cuboid(double topEnlarge, double botEnlarge, double zLength)
		{
			double[] pos = this.initialPosition;

			double z_bottom = -zLength / 2;
			double z_top = zLength / 2;

			double lenrow = zLength / numrow;
			double radius = 1;
			double angle = 2 * Math.PI / numsec;

			int indexer = 0; double r = radius;
			int ncircles = numrow + 1; double shift = -Math.PI / 4;
			for (int i = 0; i < ncircles; i++)
			{
				// positions
				radius = topEnlarge * ((ncircles - i) / (double)ncircles) +
					(i / (double)ncircles) * botEnlarge;
				for (int j = 0; j < numsec; j++)
				{
					// compute the r length
					double theta = angle * j + shift;
					double oldTheta = theta;
					int jj = (int)(theta * 4 / Math.PI);
					int kk = (jj / 2) * 2;
					theta = theta - kk * Math.PI / 4;
					theta = theta > Math.PI / 4 ? Math.PI / 2 - theta : theta;
					r = radius / Math.Cos(theta);

					// compute x,y coords
					int b = indexer * 3;
					double x = r * Math.Cos(oldTheta);
					double y = r * Math.Sin(oldTheta);
					double z = z_bottom + i * lenrow;
					pos[b] = x;
					pos[b + 1] = y;
					pos[b + 2] = z;
					indexer++;
				}
			}
			int c = indexer * 3;
			pos[c] = 0;
			pos[c+1] = 0;
			pos[c + 2] = z_bottom; // bottom vertex
			pos[c+3] = 0;
			pos[c + 4] = 0;
			pos[c + 5] = z_top; // top vertex
		}
		private void UpdateInitPosition_Cuboid
		(
			double topXEnlarge, double topYEnlarge,
			double botXEnlarge, double botYEnlarge,
			double zLength
		)
		{
			double[] pos = this.initialPosition;
			double z_bottom = -zLength / 2;
			double z_top = zLength / 2;
			double lenrow = zLength / numrow;

			int indexer = 0;
			int ncircles = numrow + 1;
			for (int i = 0; i < ncircles; i++)
			{
				// get the key points for top and bottom faces
				int shift = numsec / 4;
				double top_dx = topXEnlarge / shift, top_dy = topYEnlarge / shift;
				double bot_dx = botXEnlarge / shift, bot_dy = botYEnlarge / shift;
				double tx = topXEnlarge, ty = topYEnlarge;
				double bx = botXEnlarge, by = botYEnlarge;
				Vector2d[] Top = new Vector2d[4] {
					new Vector2d(tx, -ty), // top face,
					new Vector2d(tx, ty),
					new Vector2d(-tx, ty),
					new Vector2d(-tx, -ty)
				};
				Vector2d[] Bot = new Vector2d[4] {
					new Vector2d(bx, -by), // bottom face
					new Vector2d(bx, by),
					new Vector2d(-bx, by),
					new Vector2d(-bx, -by)
				};
				for (int k = 0; k < 4; ++k)
				{
					int l = (k + 1) % 4; double r = (ncircles - i) / (double)ncircles;
					Vector2d S = Top[k] * r + Bot[k] * (1 - r); // start point
					Vector2d T = Top[l] * r + Bot[l] * (1 - r); // end

					int s = k * shift, t = s + shift;
					for (int j = 0; j < shift; ++j)
					{
						// compute x,y coords
						int b = indexer * 3;
						double g = (double)j / shift;
						Vector2d Q = T * g + S * (1 - g);
						pos[b] = Q.x;
						pos[b + 1] = Q.y;
						pos[b + 2] = z_bottom + i * lenrow;;
						indexer++;
					}
				}
			}
			int c = indexer * 3;
			pos[c] = 0;
			pos[c + 1] = 0;
			pos[c + 2] = z_bottom; // bottom vertex
			pos[c + 3] = 0;
			pos[c + 4] = 0;
			pos[c + 5] = z_top; // top vertex
		}

		public int index;

		public Vector3d color = new Vector3d(1.0,0,0);

		public Mesh OrigMesh { get; set; }
		public Mesh FitMesh { get; set; }

	}
	public class SphereProxy : Proxy
	{
		public int level = 16;
		public SphereProxy()
		{
			Sphere sphere = new Sphere((byte)16);

			OrigMesh = sphere.GetMesh();
			FitMesh = sphere.GetMesh();

			this.type = ProxyType.Sphere;
			
			// init positions
			this.startPosition = OrigMesh.VertexPos.Clone() as double[];
			this.initialPosition = OrigMesh.VertexPos.Clone() as double[];

			// featured vurve and key points
			this.featuredCurves = new List<FeatureCurve>();
			
			FeatureCurve curve1 = new FeatureCurve(this, 0);
		//	FeatureCurve curve2 = new FeatureCurve(this, 1);
		//	FeatureCurve curve3 = new FeatureCurve(this, 2);

			int n = sphere.keyPoints.Count;
		//	int s1 = n / 3, s2 = 2*n/3;
			for (int i = 0; i < n; ++i)
			{
				int vIndex = sphere.keyPoints[i];
				KeyPoint pt = new KeyPoint(this, vIndex);
				curve1.keyPoints.Add(pt);
			}
		/*	for (int i = s1; i < s2; ++i)
			{
				KeyPoint pt = new KeyPoint(this, sphere.keyPoints[i]);
				curve2.keyPoints.Add(pt);
			}
			for (int i = s2; i < n; ++i)
			{
				KeyPoint pt = new KeyPoint(this, sphere.keyPoints[i]);
				curve3.keyPoints.Add(pt);
			}
		*/	
			this.featuredCurves.Add(curve1);
		//	this.featuredCurves.Add(curve2);
		//	this.featuredCurves.Add(curve3);
		}
	}
	public class CylinderProxy : Proxy
	{
		public CylinderProxy()
		{
			Cylinder cylinder = new Cylinder(4);
			OrigMesh = cylinder.GetMesh();
			FitMesh = cylinder.GetMesh();

			this.type = ProxyType.Cylinder;
			
			// init positions
			this.startPosition = OrigMesh.VertexPos.Clone() as double[];
			this.initialPosition = OrigMesh.VertexPos.Clone() as double[];
			this.numrow = cylinder.numrow;
			this.numsec = cylinder.numsec;

			// featured curves and key points
			this.featuredCurves = new List<FeatureCurve>();
			int n = cylinder.keyPoints.Count;
			int m = n / 2;
			FeatureCurve curve1 = new FeatureCurve(this,0);
			for (int i = 0; i < m; ++i)
			{
				KeyPoint pt = new KeyPoint(this, cylinder.keyPoints[i]);
				curve1.keyPoints.Add(pt);
			}
			FeatureCurve curve2 = new FeatureCurve(this,1);
			for (int i = m; i < n; ++i)
			{
				KeyPoint pt = new KeyPoint(this, cylinder.keyPoints[i]);
				curve2.keyPoints.Add(pt);
			}

			this.featuredCurves.Add(curve1);
			this.featuredCurves.Add(curve2);
		}
	}
	public class CuboidProxy : Proxy
	{
		public int level = -1;
		public int whichAxisTranscated = 2;
		public List<List<int>> xRects = null;
		public List<List<int>> yRects = null;
		public List<List<int>> zRects = null;
		public void GetXYZRects()
		{
			GetXRects();
			GetYRects();
			GetZRects();
		}
		public void GetZRects()
		{
			zRects = new List<List<int>>();

			List<int> top = new List<int>();
			List<int> bot = new List<int>();
			int n = (level - 1) * (level - 1);
			for (int i = 0; i < 2; ++i)
			{
				int start = i == 0 ? 0 : (numrow) * numsec;
				List<int> lst = i == 0 ? bot : top; 
				for (int j = 0; j < numsec; ++j)
				{
					lst.Add(j + start);
				}
				
				int w = i == 0 ? 0 : n;
				int t = numsec * (numrow + 1) + w;
				for (int j = 0; j < n; ++j)
				{
					lst.Add(t + j);
				}
			}

			zRects.Add(bot);
			for (int i = 1; i < numrow; ++i)
			{
				List<int> row = new List<int>();
				for (int j = 0; j < numsec; ++j)
				{
					row.Add(i * numsec + j);
				}
				zRects.Add(row);
			}
			zRects.Add(top);
		}
		public void GetXRects_old()
		{
			xRects = new List<List<int>>();

			List<int> top = new List<int>();
			List<int> bot = new List<int>();
			for (int i = 0; i < 2; ++i)
			{
				int s = i == 0 ? level*3 : level;
				List<int> lst = i == 0 ? bot : top;
				for (int k = 0; k < level+1; ++k)
				{
					for (int j = 0; j < numrow + 1; ++j)
					{
						int t = j * numsec;
						lst.Add((s + k)%numsec + t);
					}
				}
			}

			xRects.Add(bot);
			int n = (numrow+1)*numsec, m = n-numsec;
			for (int i = 1; i < level; ++i)
			{
				List<int> sec = new List<int>();
				int s = i, k = level - 1;
				for (int j = 0; j < numrow+1; ++j)
				{
					sec.Add(j * numsec + s);
				}
				s = i - 1 + n + k*k;
				for (int j = 0; j < level-1; ++j)
				{
					sec.Add(s + j * k);
				}
				s = level*3 - i + m;
				for (int j = 0; j < numrow + 1; ++j)
				{
					sec.Add(s - j * numsec);
				}
				s = k*(level-2) + n + i-1;
				for (int j = 0; j < level - 1; ++j)
				{
					sec.Add(s - j * k);
				}
				xRects.Add(sec);
			}
			xRects.Add(top);
		}
		public void GetXRects()
		{
			xRects = new List<List<int>>();

			List<int> top = new List<int>();
			List<int> bot = new List<int>();
			for (int i = 0; i < 2; ++i)
			{
				int s = i == 0 ? 0 : level;
				List<int> lst = i == 0 ? bot : top;

				int shift = s;
				for (int j = 0; j < numrow + 1; ++j)
				{
					lst.Add(j * numsec + s);
				}
				shift = i == 0 ? lst[lst.Count - 1] + numsec : lst[lst.Count - 1]; /// last corner
				if (i == 0)
				{
					for (int j = 1; j <= level; ++j)
					{
						lst.Add(shift - j);
					}
				}
				else
				{
					for (int j = 1; j <= level; ++j)
					{
						lst.Add(shift + j);
					}
				}
				shift = lst[lst.Count - 1];
				for (int j = 1; j < numrow + 1; ++j)
				{
					lst.Add(shift - j * numsec);
				}
				shift = lst[lst.Count - 1];
				if (i == 0)
				{
					for (int j = 1; j < level; ++j)
					{
						lst.Add(shift + j);
					}
				}
				else
				{
					for (int j = 1; j < level; ++j)
					{
						lst.Add(shift - j);
					}
				}
			}

			xRects.Add(bot);

			int n = (numrow + 1) * numsec, m = n - numsec;
			for (int i = 1; i < level; ++i)
			{
				List<int> sec = new List<int>();
				int s = i, k = level - 1;
				for (int j = 0; j < numrow + 1; ++j)
				{
					sec.Add(j * numsec + s);
				}
				s = i - 1 + n + k * k;
				for (int j = 0; j < level - 1; ++j)
				{
					sec.Add(s + j * k);
				}
				s = level * 3 - i + m;
				for (int j = 0; j < numrow + 1; ++j)
				{
					sec.Add(s - j * numsec);
				}
				s = k * (level - 2) + n + i - 1;
				for (int j = 0; j < level - 1; ++j)
				{
					sec.Add(s - j * k);
				}
				xRects.Add(sec);
			}

			xRects.Add(top);
		}
		public void GetYRects_old()
		{
			yRects = new List<List<int>>();

			List<int> top = new List<int>();
			List<int> bot = new List<int>();
			for (int i = 0; i < 2; ++i)
			{
				int s = i == 0 ? level * 2 : 0;
				List<int> lst = i == 0 ? bot : top;
				for (int k = 0; k < level + 1; ++k)
				{
					for (int j = 0; j < numrow + 1; ++j)
					{
						int t = j * numsec;
						lst.Add(s + k + t);
					}
				}
			}

			yRects.Add(bot);
			int n = (numrow + 1) * numsec, m = n - numsec;
			for (int i = 1; i < level; ++i)
			{
				List<int> sec = new List<int>();
				int s = level*3+i, k = level - 1;
				for (int j = 0; j < numrow + 1; ++j)
				{
					sec.Add(j * numsec + s);
				}
				s = k * (k+k-i) + n;
				for (int j = 0; j < level - 1; ++j)
				{
					sec.Add(s + j);
				}
				s = level * 2 - i + m;
				for (int j = 0; j < numrow + 1; ++j)
				{
					sec.Add(s - j * numsec);
				}
				s = n + k*(level-i)-1;
				for (int j = 0; j < level - 1; ++j)
				{
					sec.Add(s - j);
				}
				yRects.Add(sec);
			}
			yRects.Add(top);
		}
		public void GetYRects()
		{
			yRects = new List<List<int>>();

			List<int> top = new List<int>();
			List<int> bot = new List<int>();
			for (int i = 0; i < 2; ++i)
			{
				int s = i == 0 ? level*3 : 0;
				List<int> lst = i == 0 ? bot : top;

				int shift = s;
				for (int j = 0; j < numrow + 1; ++j)
				{
					lst.Add(j * numsec + s);
				}

				shift = lst[lst.Count - 1];
				if (i == 0)
				{
					for (int j = 1; j <= level; ++j)
					{
						lst.Add(shift - j);
					}
				}
				else
				{
					for (int j = 1; j <= level; ++j)
					{
						lst.Add(shift + j);
					}
				}
				
				shift = lst[lst.Count - 1];
				for (int j = 1; j < numrow + 1; ++j)
				{
					lst.Add(shift - j * numsec);
				}

				shift = lst[lst.Count - 1];
				if (i == 0)
				{
					for (int j = 1; j < level; ++j)
					{
						lst.Add(shift + j);
					}
				}
				else
				{
					for (int j = 1; j < level; ++j)
					{
						lst.Add(shift - j);
					}
				}
			}

			yRects.Add(bot);
			int n = (numrow + 1) * numsec, m = n - numsec;
			for (int i = 1; i < level; ++i)
			{
				List<int> sec = new List<int>();
				int s = level * 3 + i, k = level - 1;
				for (int j = 0; j < numrow + 1; ++j)
				{
					sec.Add(j * numsec + s);
				}
				s = k * (k + k - i) + n;
				for (int j = 0; j < level - 1; ++j)
				{
					sec.Add(s + j);
				}
				s = level * 2 - i + m;
				for (int j = 0; j < numrow + 1; ++j)
				{
					sec.Add(s - j * numsec);
				}
				s = n + k * (level - i) - 1;
				for (int j = 0; j < level - 1; ++j)
				{
					sec.Add(s - j);
				}
				yRects.Add(sec);
			}
			yRects.Add(top);
		}

		public int FindWhichAxis()
		{
			if (this.handleCurve == null) return -1;

			int id = this.featuredCurves.IndexOf(this.handleCurve);

			if (id < 2) return 2;
			if (id < 4) return 0;
			return 1;
		}

		private double[] staticPosition = null;

		public CuboidProxy()
		{
			this.level = 8;
			Cuboid cuboid = new Cuboid((byte)this.level);
			OrigMesh = cuboid.GetMesh();
			FitMesh = cuboid.GetMesh();

			this.type = ProxyType.Cuboid;

			/// init positions
			this.startPosition = OrigMesh.VertexPos.Clone() as double[];
			this.initialPosition = OrigMesh.VertexPos.Clone() as double[];
			this.numrow = cuboid.numrow;
			this.numsec = cuboid.numsec;

			/// featured curves and key points, 
			/// (0,1) --- along z axis, (2,3) --- along x axis, and (4,5) --- along y axis
			this.featuredCurves = new List<FeatureCurve>();
			int n = cuboid.zSideKeyPoints.Count;
			int m = n / 2;
			FeatureCurve curve1 = new FeatureCurve(this,0);
			for (int i = 0; i < m; ++i)
			{
				KeyPoint pt = new KeyPoint(this, cuboid.zSideKeyPoints[i]);
				curve1.keyPoints.Add(pt);
			}
			FeatureCurve curve2 = new FeatureCurve(this,1);
			for (int i = m; i < n; ++i)
			{
				KeyPoint pt = new KeyPoint(this, cuboid.zSideKeyPoints[i]);
				curve2.keyPoints.Add(pt);
			}

			/// x side feature curves
			n = cuboid.xSideKeyPoints.Count;
			m = n / 2;
			FeatureCurve curve3 = new FeatureCurve(this,2);
			for (int i = 0; i < m; ++i)
			{
				KeyPoint pt = new KeyPoint(this, cuboid.xSideKeyPoints[i]);
				curve3.keyPoints.Add(pt);
			}
			FeatureCurve curve4 = new FeatureCurve(this,3);
			for (int i = m; i < n; ++i)
			{
				KeyPoint pt = new KeyPoint(this, cuboid.xSideKeyPoints[i]);
				curve4.keyPoints.Add(pt);
			}

			/// y side feature curves
			n = cuboid.ySideKeyPoints.Count;
			m = n / 2;
			FeatureCurve curve5 = new FeatureCurve(this,4);
			for (int i = 0; i < m; ++i)
			{
				KeyPoint pt = new KeyPoint(this, cuboid.ySideKeyPoints[i]);
				curve5.keyPoints.Add(pt);
			}
			FeatureCurve curve6 = new FeatureCurve(this,5);
			for (int i = m; i < n; ++i)
			{
				KeyPoint pt = new KeyPoint(this, cuboid.ySideKeyPoints[i]);
				curve6.keyPoints.Add(pt);
			}

			this.featuredCurves.Add(curve1);
			this.featuredCurves.Add(curve2);

			this.featuredCurves.Add(curve3);
			this.featuredCurves.Add(curve4);
			this.featuredCurves.Add(curve5);
			this.featuredCurves.Add(curve6);


			this.GetXYZRects();

			this.staticPosition = this.initialPosition.Clone() as double[];
		}
		public void UpdateFeatureCurves()
		{
			if (this.onedegreeFreedom)
			{
				List<FeatureCurve> newFCurves = new List<FeatureCurve>();
				for (int i = 0; i < 2; ++i)
				{
					newFCurves.Add(this.featuredCurves[i]);
				}
				this.featuredCurves = newFCurves;
			}
		}
		public FeatureCurve FindFeatureCurve(int vertexId)
		{
			int n = FitMesh.VertexCount;
			if (vertexId < 0 || vertexId >= n)
				return null;
			
			int curveId = -1;
			if (vertexId == n - 1 || (vertexId > n-2-numsec && vertexId < n - 2))
				curveId = 1;
			else if (vertexId == n - 2 || vertexId < numsec)
				curveId = 0;
			else
			{
				int s = vertexId % this.numsec;
				int t = s / (this.numsec / 4);
				switch (t)
				{
					case 0:
						curveId = 4;
						break;
					case 1:
						curveId = 2;
						break;
					case 2:
						curveId = 5;
						break;
					case 3:
						curveId = 3;
						break;
				}
			}
			if (curveId != -1)
				return this.featuredCurves[curveId];
			return null;
		}

		public void UpdateRects(int whichAxis)
		{
			List<List<int>> rects = null;
			switch (whichAxis)
			{
				case 0: rects = this.xRects;
					break;
				case 1: rects = this.yRects;
					break;
				case 2: rects = this.zRects;
					break;
			}

			if (rects == null) return;

			int n = rects.Count; double[] pos = this.FitMesh.VertexPos;
			List<int> c0 = rects[0];
			List<int> c1 = rects[n-1];
			for (int i = 1; i < n - 1; ++i)
			{
				List<int> rect = rects[i];

				double r = (double)i / (n-1); double t = 1 - r;

				int index = 0;
				foreach (int j in rect)
				{
					int u = c0[index] * 3;
					int v = c1[index] * 3;
					int w = j * 3;

					pos[w] = t * pos[u] + r * pos[v];
					pos[w+1] = t * pos[u+1] + r * pos[v+1];
					pos[w+2] = t * pos[u+2] + r * pos[v+2];

					index++;
				}
			}
		}
		public void UpdateShape(int whichAxis)
		{
			UpdateRects(whichAxis);
			UpdateRects((whichAxis+1)%3);
			UpdateRects((whichAxis+2)%3);

			this.transcated = true;
			this.whichAxisTranscated = whichAxis;

			this.AssignKeyPointsPos();
			this.ComputeCurveInfos();
			this.UpdateScale();
			this.ComputeFeatureCurveEncodeFrames(false);
		}
		private void UpdateScale()
		{
			this.FitScale.z = (this.featuredCurves[1].centroid - this.featuredCurves[0].centroid).Length()/2;
			if (!this.onedegreeFreedom)
			{
				this.FitScale.y = (this.featuredCurves[5].centroid - this.featuredCurves[4].centroid).Length()/2;
				this.FitScale.x = (this.featuredCurves[3].centroid - this.featuredCurves[2].centroid).Length()/2;
			}
		}
		public void UpdateCurveOrientation(int whichAxis)
		{
			int i = 0, j = 1; Vector3d axis = z;
			switch (whichAxis)
			{
				case 0: i = 2; j = 3;
					break;
				case 1: i = 4; j = 5;
					break;
				case 2: i = 0; j = 1;
					break;
			}

			axis = (this.featuredCurves[j].centroid - this.featuredCurves[i].centroid).Normalize();

			this.featuredCurves[i].UpdateMeshPos();
			this.featuredCurves[j].UpdateMeshPos();

			/// update the axis of the proxy
			HelperFunction.UpdateProxyAxis(this, axis, whichAxis);

			double[] pos = this.FitMesh.VertexPos;
			for (int k = 0; k < 2; ++k)
			{
				FeatureCurve c = k == 0 ? this.featuredCurves[i] : this.featuredCurves[j];
				double cos = c.frame.N.Normalize().Dot(axis);
				if (cos < 0)
				{
					axis = new Vector3d() - axis;
				}

				cos = Math.Abs(cos);

				if (cos > 1) cos = 1;

			///	if (cos > ProxyPropagator2.cos3) 
			///		continue;

				Vector3d rotAxis = c.frame.N.Cross(axis).Normalize();
				if (double.IsNaN(rotAxis.x))
					return;

				double rotAngle = Math.Acos(Math.Abs(cos));
				Matrix4d R = Matrix4d.RotationMatrix(rotAxis, rotAngle);

				Vector3d center = c.centroid;
				foreach (KeyPoint kpt in c.keyPoints)
				{
					int b = kpt.index * 3;
					Vector3d p = kpt.pos;
					Vector3d q = (R * new Vector4d(p - center, 0)).XYZ() + center;

					kpt.pos = q;

					pos[b] = q.x;
					pos[b+1] = q.y;
					pos[b+2] = q.z;
				}
			}
		}
	}

	public unsafe class GeneralizedCylinderProxy : Proxy
	{
		public double createdRadius = 1;
		public ControlCircle[] controlCircles = null;
		public UserCurve cubicSpline = null;
		public Matrix4d topTransform = null;
		public Matrix4d botTransform = null;
		public GeneralizedCylinderProxy(Mesh m)
		{
			OrigMesh = m;
			FitMesh = new Mesh(m);

			this.type = ProxyType.GeneralizedCylinder;

			/// init positions & encoding frames
			this.UpdateEncodingFrame();
			this.startPosition = OrigMesh.VertexPos.Clone() as double[];
			this.initialPosition = OrigMesh.VertexPos.Clone() as double[];

		}
		public void UpdateRaiuds(double ratio)
		{
			double scal = this.controlCircles[0].scale * ratio;
			for (int i = 0; i < this.controlCircles.Length; ++i)
			{
				ControlCircle c = this.controlCircles[i];
				double s = c.scale * ratio;
				c.scale = s;

				Matrix4d S = Matrix4d.ScalingMatrix(ratio, ratio, 1);
				Matrix4d R = new Matrix4d(
					new Matrix3d(
						c.xAxis,
						c.yAxis,
						c.zAxis
					).Transpose()
				);

				Matrix4d Q = R.Inverse() * S * R;

				Vector3d o = c.center;

			//	double[] orgpos = OrigMesh.VertexPos;
				double[] pos = FitMesh.VertexPos;
				foreach (int v in c.vertexIndexes)
				{
					int b = v * 3;
					Vector3d p = new Vector3d(pos, b);
					Vector3d q = (Q * new Vector4d(p-o, 0)).XYZ() + o;
					pos[b] = q.x;
					pos[b + 1] = q.y;
					pos[b + 2] = q.z;
				}
			}

			this.FitScale.x = this.FitScale.y = scal;
			this.featuredCurves[0].radius = scal;
			this.featuredCurves[0].scale.x = this.featuredCurves[0].scale.y = scal;
			this.featuredCurves[1].radius = scal;
			this.featuredCurves[1].scale.x = this.featuredCurves[1].scale.y = scal;
		}
		public double GetVolume()
		{
			int n = this.controlCircles.Length;
			double zLength = 0;
			for (int i = 0; i < n - 1; ++i)
			{
				Vector3d d = this.controlCircles[i + 1].center - 
					this.controlCircles[i].center;
				zLength += d.Length();
			}
			double radius = this.FitScale.x;
			return zLength * Math.PI * radius * radius;
		}
		public void SetKeyPoints()
		{
			if (this.controlCircles == null)
			{
				Program.OutputText("Fail to set key points ...",true);
				return;
			}
			
			int n = this.controlCircles.Length;

			ControlCircle topCircle = this.controlCircles[0];
			ControlCircle botCircle = this.controlCircles[n-1];

			this.featuredCurves = new List<FeatureCurve>();
			FeatureCurve curve1 = new FeatureCurve(this,0);
			for (int i = 0; i < topCircle.vertexIndexes.Count-1; ++i)
			{
				int v = topCircle.vertexIndexes[i];
				KeyPoint kpt = new KeyPoint(this, v);
				curve1.keyPoints.Add(kpt);
			}
			FeatureCurve curve2 = new FeatureCurve(this,1);
			for (int i = 0; i < botCircle.vertexIndexes.Count - 1; ++i)
			{
				int v = botCircle.vertexIndexes[i];
				KeyPoint kpt = new KeyPoint(this, v);
				curve2.keyPoints.Add(kpt);
			}
			this.featuredCurves.Add(curve1);
			this.featuredCurves.Add(curve2);

			this.AssignKeyPointsPos();
		}
		public void HandleUpdate(Matrix4d T)
		{
			if (this.optimized) return;
			
			double[] orgpos = OrigMesh.VertexPos;
			double[] pos = FitMesh.VertexPos;

			/// solve new positions
			Matrix4d I = Matrix4d.IdentityMatrix();
			Matrix4d Q1 = this.handleCurve.inProxyIndex == 0 ? T : I;
			Matrix4d Q2 = this.handleCurve.inProxyIndex == 0 ? I : T;

			HandleUpdate(Q1,Q2);

			this.UpdateEncodingFrame();
			this.optimized = true;
		}
		public void HandleUpdate(Matrix4d Q1, Matrix4d Q2)
		{
			if (this.optimized) return;
			this.optimized = true;

			double[] orgpos = OrigMesh.VertexPos;
			double[] pos = FitMesh.VertexPos;

			/// solve new positions
			int K = this.controlCircles.Length;
			for (int i = 0; i < K; ++i)
			{
				double r = (double)i / (K - 1);
				Matrix4d Q = Q1 * (1 - r) + Q2 * r;
				ControlCircle circle = this.controlCircles[i];
				Vector3d c = circle.oldCenter;
				foreach (int v in circle.vertexIndexes)
				{
					int b = v * 3;
					Vector3d p = new Vector3d(orgpos, b);
					Vector3d q = (Q * new Vector4d(p - c, 1)).XYZ() + c;
					pos[b] = q.x;
					pos[b + 1] = q.y;
					pos[b + 2] = q.z;
				}

				circle.xAxis = (Q * new Vector4d(circle.xAxis, 0)).XYZ();
				circle.yAxis = (Q * new Vector4d(circle.yAxis, 0)).XYZ();
				circle.zAxis = (Q * new Vector4d(circle.zAxis, 0)).XYZ();

				circle.center = (Q * new Vector4d(new Vector3d(), 1)).XYZ() + c;
			}
			this.AssignKeyPointsPos();
			this.ResetLocalFrames();
			this.ComputeFeatureCurveEncodeFrames(true);
		}
		public void UserPropagateUpdate()
		{
			if (this.optimized) return;
			this.optimized = true;

			/// first find top-bottom scales, new orientatio, and translation
			double[] pos = this.FitMesh.VertexPos;
			double[] orgpos = this.OrigMesh.VertexPos;

			int n = this.featuredCurves[0].keyPoints.Count;
			int K = this.controlCircles.Length;

			ControlCircle topCircle = this.controlCircles[0];
			ControlCircle botCircle = this.controlCircles[K - 1];

			FeatureCurve top = this.featuredCurves[0];
			FeatureCurve bot = this.featuredCurves[1];

			/// new centers
			Vector3d c1 = top.centroid;
			Vector3d c2 = bot.centroid;

			Vector3d oldc1 = topCircle.oldCenter;
			Vector3d oldc2 = botCircle.oldCenter;

			Matrix4d T1 = Matrix4d.TranslationMatrix(c1 - oldc1); // 1 translation
			Matrix4d T2 = Matrix4d.TranslationMatrix(c2 - oldc2);

			/// new scales
			double radiusTop = top.radius, radiusBot = bot.radius;

			/// whether allow one side enlarge or not
			double s1 = radiusTop, s2 = radiusBot;
			if (!this.AllowAxisEnlarge)
			{
				s1 = s2 = (radiusTop + radiusBot) / 2;
			}

			double olds1 = this.FitScale.x;
			double olds2 = this.FitScale.y;
			s1 /= olds1;
			s2 /= olds2;

			Matrix4d S1 = Matrix4d.ScalingMatrix(s1, s1, s1);
			Matrix4d S2 = Matrix4d.ScalingMatrix(s2, s2, s2);

			/// find top-bot rotations
			Vector3d n1 = top.frame.N;
			Vector3d n2 = bot.frame.N;

			Vector3d oldn1 = topCircle.zAxis;
			Vector3d oldn2 = botCircle.zAxis;

			Matrix4d I = Matrix4d.IdentityMatrix();
			Matrix4d R1 = I;
			Matrix4d R2 = I;
			if (!double.IsNaN(n1.x))
			{
				double cos = oldn1.Dot(n1);
				if (cos < 0) n1 = new Vector3d() - n1;
				double cos1 = Math.Abs(cos);
				if (cos1 < Math.Cos(1.0 / 18 * Math.PI)) // > 10 degree
				{
					Vector3d rotAxis = oldn1.Cross(n1).Normalize();
					double angle = Math.Acos(cos1);
					R1 = Matrix4d.RotationMatrix(
						rotAxis,
						angle
					);
				}
			}
			if (!double.IsNaN(n2.x))
			{
				double cos = oldn2.Dot(n2);
				if (cos < 0) n2 = new Vector3d() - n2;
				double cos1 = Math.Abs(cos);
				if (cos1 < Math.Cos(1.0 / 18 * Math.PI)) // > 10 degree
				{
					Vector3d rotAxis = oldn2.Cross(n2).Normalize();
					double angle = Math.Acos(cos1);
					R2 = Matrix4d.RotationMatrix(
						rotAxis,
						angle
					);
				}
			}

			/// solve new positions
			Matrix4d Q1 = T1;
			Matrix4d Q2 = T2;
			for (int i = 0; i < K; ++i)
			{
				ControlCircle circle = this.controlCircles[i];
				Matrix4d O = new Matrix4d(
					new Matrix3d(
						circle.xAxis,
						circle.yAxis,
						circle.zAxis
					).Transpose()
				);

				Matrix4d S3 = R1 * O.Inverse() * S1 * O;
				Matrix4d S4 = R2 * O.Inverse() * S2 * O;

				double r = (double)i / (K-1);
				Matrix4d Q = S3 * (1 - r) + S4 * r;

				/// scaling & rotation
				Vector3d c = circle.oldCenter;
				foreach (int v in circle.vertexIndexes)
				{
					int b = v * 3;
					Vector3d p = new Vector3d(orgpos, b);
					Vector3d q = (Q * new Vector4d(p - c, 0)).XYZ() + c;
					pos[b] = q.x;
					pos[b + 1] = q.y;
					pos[b + 2] = q.z;
				}

				/// translation
				Q = Q1 * (1 - r) + Q2 * r;
				foreach (int v in circle.vertexIndexes)
				{
					int b = v * 3;
					Vector3d p = new Vector3d(pos, b);
					Vector3d q = (Q * new Vector4d(p, 1)).XYZ();
					pos[b] = q.x;
					pos[b + 1] = q.y;
					pos[b + 2] = q.z;
				}

				circle.xAxis = (Q * new Vector4d(circle.xAxis, 0)).XYZ();
				circle.yAxis = (Q * new Vector4d(circle.yAxis, 0)).XYZ();
				circle.zAxis = (Q * new Vector4d(circle.zAxis, 0)).XYZ();

				circle.center = (Q * new Vector4d(new Vector3d(), 1)).XYZ() + c;
			}

			topCircle.zAxis = new Vector3d(n1);
			botCircle.zAxis = new Vector3d(n2);

			this.AssignKeyPointsPos();
			this.ResetLocalFrames();
			this.ComputeFeatureCurveEncodeFrames(true);
			this.UpdateEncodingFrame();
		}
		public void UpdateOldCenters()
		{
			foreach (ControlCircle c in this.controlCircles)
			{
				c.oldCenter = new Vector3d(c.center);
			}
		}
		public void UpdateControlCenters(Matrix4d T)
		{
			foreach (ControlCircle c in this.controlCircles)
			{
				c.center = (T * (new Vector4d(c.oldCenter, 1))).XYZ();
			}
		}
		public void Fit2Curve(UserCurve target)
		{
			if (target == null || target.numofPoints != this.controlCircles.Length)
				return;

			double[] orgpos = OrigMesh.VertexPos;
			double[] pos = FitMesh.VertexPos;

			/// solve new positions
			int K = this.controlCircles.Length;
			for (int i = 0; i < K; ++i)
			{
				ControlCircle circle = this.controlCircles[i];

				Vector3d c = circle.center;
				Vector3d z = circle.zAxis;

				Vector3d t = target.samplePoints[i] - c;
				Vector3d n = target.pointsDerivatives[i];

				Matrix4d T = Matrix4d.TranslationMatrix(t);
				Matrix4d R = Matrix4d.IdentityMatrix();

				double cos = Math.Abs(z.Dot(n) / z.Length() / n.Length());
				if (cos < HelperFunction.cos5)
				{
					double rotAngle = Math.Acos(cos);
					Vector3d rotAxis = z.Cross(n).Normalize();
					R = Matrix4d.RotationMatrix(rotAxis, rotAngle);
				}

				Matrix4d Q = T * R;

				Vector3d o = circle.oldCenter;
				foreach (int v in circle.vertexIndexes)
				{
					int b = v * 3;
					Vector3d p = new Vector3d(orgpos, b);
					Vector3d q = (Q * new Vector4d(p - o, 1)).XYZ() + o;
					pos[b] = q.x;
					pos[b + 1] = q.y;
					pos[b + 2] = q.z;
				}
				circle.center = (Q * new Vector4d(new Vector3d(), 1)).XYZ() + o;
				circle.zAxis = (Q * new Vector4d(circle.zAxis, 0)).XYZ().Normalize();
				circle.xAxis = (Q * new Vector4d(circle.xAxis, 0)).XYZ().Normalize();
				circle.yAxis = circle.zAxis.Cross(circle.xAxis).Normalize();
				circle.xAxis = circle.yAxis.Cross(circle.zAxis).Normalize();

				this.optimized = true;
			}

			this.UpdateControlCurveFrame();

			this.ComputeFeatureCurvesCentroid();
			this.ComputeFeatureCurvesRadius();
			this.ComputeFeatureCurvesFrame();

			this.ResetLocalFrames();
			this.ComputeFeatureCurveEncodeFrames(true);

		}
		public void CreateControlCurves()
		{
			if (this.controlCurves == null) this.controlCurves = new List<FeatureCurve>();
			foreach (ControlCircle cirlce in this.controlCircles)
			{
				FeatureCurve c = new FeatureCurve(this, -1);
				foreach (int v in cirlce.vertexIndexes)
				{
					KeyPoint kpt = new KeyPoint(this, v);
					c.keyPoints.Add(kpt);
				}
				c.frame = new LocalFrame();
				c.frame.b1 = cirlce.xAxis;
				c.frame.b2 = cirlce.yAxis;
				c.frame.N = cirlce.zAxis;
				c.centroid = cirlce.center;
				c.oldCentroid = new Vector3d(c.centroid);
				c.controlFrame = new LocalFrame[cirlce.vertexIndexes.Count];
				for (int i = 0; i < cirlce.vertexIndexes.Count; ++i)
				{
					c.controlFrame[i] = new LocalFrame();
				}
				c.UpdateControlFrame();

				this.controlCurves.Add(c);
			}
		}
		public void UpdateControlCurveFrame()
		{
			if (this.controlCurves.Count != this.controlCircles.Length) return;

			int index = 0;
			foreach (FeatureCurve c in this.controlCurves)
			{
				c.frame.b1 = this.controlCircles[index].xAxis;
				c.frame.b2 = this.controlCircles[index].yAxis;
				c.frame.N = this.controlCircles[index++].zAxis;
				c.UpdateControlFrame();
			}
		}
		public void UpdateCenter()
		{
			int n = this.controlCircles.Length;
			Vector3d c = this.controlCircles[n / 2].center;
			this.FitTranslate = new Vector3d(c);
			this.o = new Vector3d(c);
		}
	}


	/// class ProxyPair, for symmetry propagation only
	public class ProxyPair : PriorityQueueElement
	{
		public Proxy p = null;
		public Proxy q = null;
		public double distance = 0;
		public ProxyPair(Proxy i, Proxy j, double dis)
		{
			this.p = i;
			this.q = j;
			this.distance = dis;
		}
		#region PriorityQueueElement Members
		private int pqIndex = -1;
		public int PQIndex
		{
			get
			{
				return pqIndex;
			}
			set
			{
				pqIndex = value;
			}
		}
		public int CompareTo(PriorityQueueElement obj)
		{
			ProxyPair rec = obj as ProxyPair;
			if (this.distance < rec.distance) return -1;
			if (this.distance > rec.distance) return 1;
			return 0;
		}

		#endregion
	}
	
	/// calss proxy group
	public enum GroupStyle
	{
		Symmetry = 0,
		Parallel,
		Coplanar,
		Concentric,
		None
	}
	public class ProxyGroup
	{
		public GroupStyle property = GroupStyle.None;
		public List<Proxy> members = new List<Proxy>(); // all memebers belong to this type.
		public bool treated = false;
		public Vector3d center = new Vector3d();
		public ProxyGroup(GroupStyle prop)
		{
			this.property = prop;
		}
		public ProxyGroup()
		{
		}
	}
}
